package sfm.dataset;

import java.io.Serializable;
import java.util.*;

/**
 * Dataset hierarchy:
 * 
 *    Dataset--
 * 	          |
 *            -- Instance--
 *                        |
 *                        -- Attribute
 */
public final class Dataset implements Serializable
{
  private String[] classNames;
  private String datasetName;
  private Vector<Instance> dataset;
  private int numberOfClasses;
  
  private Vector<Instance> dsTrain, dsVal, dsTest;
  
  public Dataset(String name, int nClasses)
  {
    datasetName        = name;
    numberOfClasses    = nClasses;
    dataset            = new Vector<Instance>();
  }
//---------------------------------------------------------------------//
  public String getName()
  {
    return this.datasetName;
  }
//---------------------------------------------------------------------//
  /**
   * Sets (and replaces) the existing set of instances
   * already stored here, if any.
   * 
   * !PS!: This method may NOT be compatible with the existing
   * data-analyzing and preparation code! The instances supplied
   * here are supposed to be self-contained with regard to any (meta-)info
   * attatched to instances when loaded the normal way. 
   * (=> as when loading a dataset from file). 
   */
  public void setInstances(Vector<Instance> newInstances)
  {
    dataset = newInstances;
  }
//---------------------------------------------------------------------//
  public void addInstance(Instance instance)
  {
    dataset.addElement(instance);
  }
//---------------------------------------------------------------------//
  /**
   * returns the data as a set of instances; a list of class Instance objects
   */
  public final Vector<Instance> getInstanceData()
  {
    return dataset;
  }
//---------------------------------------------------------------------//
  /**
   * returns a 'list-of-lists' with instance data grouped by their classifications;
   * used for picking out a 'reasonably representative' set of training data...
   */
  private final Vector<Instance>[] getDataGroupedByClass()
  {
    Vector<Instance>[] v = new Vector[numberOfClasses];    
    Instance inst;
    
    for (int i=0; i < v.length; i++) v[i] = new Vector<Instance>(); 
    
    for (int i=0; i < dataset.size(); i++)
    {
      inst = dataset.elementAt(i);
      v[inst.getClassID()].addElement(inst);
    }    
    return v;
  }
//---------------------------------------------------------------------//
  /**
   * Testing-method ensuring same sets of instances
   * when developing the adabooster... (for each instantiation of it)
   */
  public void splitDatasetAndStorePartitions(double pctTrain, double pctVal)
  {
  	dsTrain = new Vector<Instance>();
  	dsVal   = new Vector<Instance>();
  	dsTest  = new Vector<Instance>();
  	
  	splitDataset(pctTrain, pctVal, dsTrain, dsVal, dsTest);
  }
//--------------------------------------------------------------------------------//
  public void getPartitionedDatasets(Vector<Instance> data_train, Vector<Instance> data_validation, Vector<Instance> data_test)
  {
  	for (int i=0; i < dsTrain.size(); i++) data_train.addElement(dsTrain.elementAt(i));
  	for (int i=0; i < dsVal.size(); i++)   data_validation.addElement(dsVal.elementAt(i));
  	for (int i=0; i < dsTest.size(); i++)  data_test.addElement(dsTest.elementAt(i));
  }
//--------------------------------------------------------------------------------//
  /**
   * While the selection of training-samples is randomized to ensure fair distribution(?), we
   * also SHOULD ensure that this will cover some instances from all CLASSES.
   *   - (I.e. the general assumption that a given dataset for training must be 'reasonably'
   * representative of the corresponding (un)limited set of real data, out there...)
   * 
   * NB: The rough method here will cause nearly all samples of some classes to be included 
   * in the training set..., if a class has very few samples relative to the total amount of
   * training samples.
   */
  private void splitDataset(double pctForTraining, double pctForValidation, Vector<Instance> data_train, Vector<Instance> data_validation, Vector<Instance> data_test)
  {
    data_train.removeAllElements();
    data_test.removeAllElements();
    if (data_validation != null) data_validation.removeAllElements();
    
    int nTrainElems      = (int) (pctForTraining   * getNumberOfInstances());
    int nValidationElems = (int) (pctForValidation * nTrainElems);
    int tmpElem;
    
    // if dataset too small, use ALL samples for training... (probably just a dummy-dataset)
    if (pctForTraining >= 1.0 || getNumberOfInstances() < 25 || nValidationElems <= 1)
    {
      pctForTraining   = 1.0;
      pctForValidation = 0.0;
      nTrainElems      = getNumberOfInstances();
    }
    
    // these are *copies* of references, so we may safely (non-destructively) remove elements
    // from these vectors, in vArr[], causing no change to the original dataset... :)
    Vector<Instance>[] vArr = getDataGroupedByClass();
    boolean progress = true;
    
    while (data_train.size() < nTrainElems && progress)
    {
      progress = false;
      
      for (int i=0; i < vArr.length; i++)
      {
        // don't spend it all on training; save at least 2 for testing
        if (vArr[i].size() > 2)
        {
          tmpElem = (int) (Math.random() * vArr[i].size());
          data_train.addElement(vArr[i].remove(tmpElem));
          progress = true;
          if (data_train.size() >= nTrainElems) break;
        }        
      }
    }
    
    // just add the any samples to satisy the size of training-set...
    while (data_train.size() < nTrainElems)
    {
      for (int i=0; i < vArr.length; i++)
      {
        // don't spend it all on training; save at least 2 for testing
        if (!vArr[i].isEmpty())
        {
          tmpElem = (int) (Math.random() * vArr[i].size());
          data_train.addElement(vArr[i].remove(tmpElem));
          if (data_train.size() >= nTrainElems) break;
        }        
      }
    }
    
    ////////////////////////////////////////////////////////////
    ///// Use the given ratio to get samples to use for the  
    ///// validation-set: these are taken from the TRAINING DATA!
    if (nValidationElems > 0 && data_validation != null)
    {
      for (int i=0; i < nValidationElems; i++)
      {
        data_validation.addElement(data_train.remove(0));
      }
    }
    
    ////////////////////////////////////////////////////////
    ///// Now put the remaining samples into the 'test' bag
    boolean empty = false;
    
    while (!empty)
    {
      empty = true;
      
      for (int i=0; i < vArr.length; i++)
      {
        if (!vArr[i].isEmpty()) 
        {
          empty = false;          
          for (int j=0; j < vArr[i].size(); j++)
          {
            data_test.addElement(vArr[i].elementAt(j));
          }
          vArr[i].removeAllElements();
        }        
      }
    }
  }
//---------------------------------------------------------------------//
  public final void setClassNames(String[] cNames)
  {
    this.classNames = cNames;
  }
//---------------------------------------------------------------------//
  public String getClassName(int classID)
  {
    return classNames[classID];
  }
//---------------------------------------------------------------------//
  /**
   * returns the size of the dataset; the number of instances
   */
  public int getNumberOfInstances() 
  {
    return dataset.size(); 
  }
//---------------------------------------------------------------------//
  public int getNumberOfClasses() 
  { 
    return numberOfClasses; 
  }
//---------------------------------------------------------------------//
}