package sfm.shp.network;

/**
 *  NodeFunction:  (1 / (1 + (Math.exp(-netInput))))  value-range:
 *   +36.7368006  == 1.0 output, 5 == 0.99
 *   0.0          == 0.5
 *   -36.7368006 == 1.1102229909599741E-16,  -5 = 0.0066
 */

import java.io.*;
import java.util.*;

import sfm.dataset.*;
import sfm.shp.gui.*;

public final class FNN implements Runnable, Serializable
{
  private Vector<Neuron[]> layerVector;
  
  private Neuron[] inputLayer, hiddenLayer, outputLayer;
  private Random rnd;
  
  private Vector<Instance> patterns_train, patterns_test, patterns_valset;
  private Dataset dataset;
  
  /**
   * -= 'min_classificationStrength' =-
   * 
   * this determines what value will be accepted as a 'correct' output
   * when '1.0' is the target value: 
   *  - if set to 0.85, it means an output of a value >= 0.85 is accepted as ok.
   *  
   *  In a net with MORE THAN one outputNode, the node with the highest output
   *  is by definition the winner. If there is only ONE outputNode, an output of
   *  0.85+ is considered a positive classification of the instance, and a value
   *  equal-or-less than (1 - 0.85) to be a definitive negative one...
   */ 
  private double min_classificationStrength, maxMappingError;  
  private double learningRate, momentum, noise;
  private int classifiedAs, nInputNodes, nHiddenNodes, nOutputNodes, epochs, nSuccesses;
  private int realInputNodes, preStructuringEpochs;
  private ANNConfig annCfg;
  private GUI_FNN gui;
  
  private double[] inputLayerNF, hiddenLayerNF;
//------------------------------------------------------------//
  /**
   * Constructor for a simple 3-layer feedforward 
   * neural network; input-, output- and 1 hidden layer.
   */
  public FNN(GUI_FNN gooey, ANNConfig annConfig)
  {
    gui    = gooey;
    annCfg = annConfig;
    loadConfig();
    
    inputLayer  = new Neuron[nInputNodes];
    hiddenLayer = new Neuron[nHiddenNodes];
    outputLayer = new Neuron[nOutputNodes];
    
    for (int i=0; i < nInputNodes;  i++) inputLayer[i]  = new Neuron(0, nHiddenNodes);
    for (int i=0; i < nHiddenNodes; i++) hiddenLayer[i] = new Neuron(nInputNodes, nOutputNodes);
    for (int i=0; i < nOutputNodes; i++) outputLayer[i] = new Neuron(nHiddenNodes, 0);

    rnd = new Random();
    initializeNetwork();
    
    layerVector = new Vector<Neuron[]>();
    layerVector.addElement(inputLayer);
    layerVector.addElement(hiddenLayer);
    layerVector.addElement(outputLayer);
    
  }
//-----------------------------------------------------------------//
  private void loadSimpleParameters()
  {
    epochs        = annCfg.getEpochs();
    learningRate  = annCfg.getLearningR();
    momentum      = annCfg.getMomentum();
    noise         = annCfg.getNoise();
    min_classificationStrength = annCfg.getMargin();
    preStructuringEpochs = annCfg.getPrestructuringEpochs();
  }
//-----------------------------------------------------------------//
  private void loadConfig()
  {
    loadSimpleParameters();
    
    nInputNodes   = annCfg.getInputNodes();
    nHiddenNodes  = annCfg.getHiddenNodes();
    nOutputNodes  = annCfg.getOutputNodes();

    patterns_train  = new Vector<Instance>();
    patterns_valset = new Vector<Instance>();
    patterns_test   = new Vector<Instance>();
    
    inputLayerNF   = new double[nInputNodes];
    hiddenLayerNF  = new double[nHiddenNodes];
    realInputNodes = (nInputNodes - nHiddenNodes);

    dataset = annCfg.getDataset();
    dataset.splitDatasetAndStorePartitions(annCfg.getTrainingRatio(), 0.0);
    dataset.getPartitionedDatasets(patterns_train, patterns_valset, patterns_test);
  }
//-----------------------------------------------------------------//
  public void setDataset(sfm.dataset.Dataset ds)
  {
    this.dataset = ds;
    
    dataset.splitDatasetAndStorePartitions(annCfg.getTrainingRatio(), 0.0);
    dataset.getPartitionedDatasets(patterns_train, patterns_valset, patterns_test);
  }
//-----------------------------------------------------------------//
  public void saveMe(String toFile) // Silje...
  {
    try
    {
      FileOutputStream fs    = new FileOutputStream(toFile + ".net");
      ObjectOutputStream out = new ObjectOutputStream(fs); 
      
      out.writeObject(annCfg);
      out.writeObject(inputLayer);
      out.writeObject(hiddenLayer);
      out.writeObject(outputLayer);
      
      out.flush();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
//-----------------------------------------------------------------//  
  public ANNConfig loadMe(String fromFile)
  {
    try
    {
      FileInputStream fs   = new FileInputStream(fromFile + ".net");
      ObjectInputStream in = new ObjectInputStream(fs); 
      
      annCfg      = (ANNConfig) in.readObject();
      inputLayer  = (Neuron[])  in.readObject();
      hiddenLayer = (Neuron[])  in.readObject();
      outputLayer = (Neuron[])  in.readObject();
      
      in.close();
      loadConfig(); // very crucial :)
      
      return annCfg;
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    catch (ClassNotFoundException e)
    {
      e.printStackTrace();
    }
    return null;
  }
//-----------------------------------------------------------------//
  private void initializeNetwork()
  {
    Link link;
    double weight;
    int flag;
    
    ///////////////////////////////////////////////
    /// FROM INPUT to HIDDEN LAYER    
    for (int i=0; i < inputLayer.length;  i++) 
    {
      for (int j=0; j < hiddenLayer.length; j++) 
      {
        flag = 1;
        if (Math.random() >= 0.5) flag = -1;
        
        weight = rnd.nextDouble() * flag;        
        // weight = (rnd.nextDouble() * 2) - 1;
        link   = new Link(weight, false);
        
        link.setInNeuron(inputLayer[i]);
        link.setOutNeuron(hiddenLayer[j]);
        
        inputLayer[i].setForwardLink(j, link);        
        hiddenLayer[j].setBackwardLink(i, link);
      }
    }
    
    ///////////////////////////////////////////////
    /// FROM HIDDEN to OUTPUT LAYER
    rnd.setSeed(System.nanoTime());
    
    for (int i=0; i < hiddenLayer.length;  i++) 
    {
      for (int j=0; j < outputLayer.length; j++) 
      {
        weight = (rnd.nextDouble() * 2) - 1;
        link   = new Link(weight, false);
        
        link.setInNeuron(hiddenLayer[i]);
        link.setOutNeuron(outputLayer[j]);
        
        hiddenLayer[i].setForwardLink(j, link);
        outputLayer[j].setBackwardLink(i, link);       
      }
    }
  }
//-----------------------------------------------------------------//
  /**
   * resets the input in the neurons to 0.0
   * Does not affect the INPUTLAYER, as it will implicitly be
   * reset when given a new input. 
   */
  private final void resetInput_HO()
  {
    for (int i=0; i < hiddenLayer.length; i++) hiddenLayer[i].setInput(0.0);
    for (int i=0; i < outputLayer.length; i++) outputLayer[i].setInput(0.0);
  }
//-----------------------------------------------------------------//
  private final void propagateInput()
  {
    for (int i=0; i < inputLayer.length; i++)  inputLayer[i].propagateForward();
    for (int i=0; i < hiddenLayer.length; i++) hiddenLayer[i].propagateForward();
  }
//-----------------------------------------------------------------//
  public final void runThisInputPattern(double[] pattern)
  {
    resetInput_HO();
    
    for (int i=0; i < pattern.length; i++)
    {
      inputLayer[i].setInput(pattern[i]);
    }
    propagateInput();
  }
//-----------------------------------------------------------------//
  private final void runThisInput(Instance inst)
  {
    runThisInputPattern(inst.getPlainNeuralNetData());
  }
//-----------------------------------------------------------------//
  /** 
   *  NB: Assuming FULL CONNECTIVITY, so that every node
   *  connects to ALL nodes in prev and/or next layer!
   */
  private final void computeInputAndHiddenLayerNFs()
  {
    for (int j=0; j < outputLayer[0].backwardConnections.length; j++)
    {
      hiddenLayerNF[j] = hiddenLayer[j].nodeFunction();
    }
    
    for (int j=0; j < hiddenLayer[0].backwardConnections.length; j++)
    {
      inputLayerNF[j] = inputLayer[j].nodeFunction();
    }
  }
//-----------------------------------------------------------------//
  private double[] getCorrectOutput(int classID)
  {
    return null;
  }
//-----------------------------------------------------------------//
  private final double backPropagate(Instance curPat)
  {
    double wChange, errorK = 0, errorJ = 0, outputK, outputJ, patternError = 0.0;
    double d, momentumAdd;
    int j, i;
    
    computeInputAndHiddenLayerNFs();
    
    double[] correctOutput = getCorrectOutput(curPat.getClassID());
    
    // for each outputNode to hiddenNodes
    for (i=0; i < outputLayer.length; i++)
    {
      // if (i == correct) d = 1.0f; else d = 0.0; Not for this kindatraining...
      
      d = correctOutput[i];
      
      outputK = outputLayer[i].nodeFunction();
      errorK  =  outputK * (1.0f - outputK) * (d - outputK);
      
      // patternError += ((d - outputK) * (d - outputK));
      patternError += Math.abs(d - outputK);
      
      outputLayer[i].tmpError = errorK;
      
      // for each backwardlink in outputNode[i]... update the linkWeight back to hiddenLayer-nodes
      for (j=0; j < outputLayer[i].backwardConnections.length; j++)
      {
        // wChange = (learningRate * errorK * hiddenLayer[j].nodeFunction()); // the pure weight change
        wChange = (learningRate * errorK * hiddenLayerNF[j]); // the pure weight change
        
        momentumAdd = (momentum * outputLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff! 
        outputLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    
    // for each hiddenNode...to inputnodes
    for (i=0; i < hiddenLayer.length; i++)
    {
      outputJ = hiddenLayer[i].nodeFunction();
      errorJ  = (outputJ * (1 - outputJ)) * getSumOfWeightedErrors(hiddenLayer[i]);      
      
      for (j=0; j < hiddenLayer[i].backwardConnections.length; j++)
      {
        // wChange = (learningRate * (errorJ * inputLayer[j].nodeFunction())); // the pure weight change
        wChange = (learningRate * (errorJ * inputLayerNF[j])); // the pure weight change
        
        momentumAdd = (momentum * hiddenLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff!
        hiddenLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    return patternError;
  }
//-----------------------------------------------------------------//
  /**
   * implicit, usable only for the hiddenlayer - and only ONE hiddenlayer...
   * The previous layer must be the outputLayer 
   */
  private final double getSumOfWeightedErrors(Neuron n)
  {
    double error = 0.0f;
    
    for (int i=0; i < outputLayer.length; i++)
    {
      error += (n.forwardConnections[i].getWeight() * outputLayer[i].tmpError);
    }   
    return error;
  }
//--------------------------------------------------------------------//
  private void startTrainingX()
  {
    gui.clearOutput();
    rnd.setSeed(System.currentTimeMillis());

    double epochError;
    int patternNo;
    Instance inst;
    
    Vector<Instance> tmpPatterns = new Vector<Instance>();
    Vector<Double> stats         = new Vector<Double>();

    long startS = System.currentTimeMillis();
    ///////////////////////////////////////////////////////////
    for (int i=0; i < epochs; i++)
    {
      gui.showProgress(i+1);
      
      ////////////////////////////////////
      /// ADD random pattern-order! HERE:
      for (int j=0; j < patterns_train.size(); j++) tmpPatterns.addElement(patterns_train.elementAt(j));
      epochError = 0;
      
      while (!tmpPatterns.isEmpty())
      {
        patternNo = rnd.nextInt(tmpPatterns.size());
        inst = tmpPatterns.elementAt(patternNo);
        
        runThisInput(inst);        
        epochError += backPropagate(inst);
        
        tmpPatterns.remove(patternNo);
        
        if (annCfg.updateWeightsPerPattern()) updateWeights();
      }

      if (annCfg.updateWeightsPerEpoch()) updateWeights();
      
      stats.addElement(new Double(epochError / patterns_train.size()));       
      // if (isFinished()) break;
    }
    /////////////////////////////////////////////////////////     
    long endS = System.currentTimeMillis();
    
    printResultsX(((endS - startS) / 1000), stats, patterns_train);     
  }
//--------------------------------------------------------------------//
  /**
   * Depending on the weight-updating scheme, updates the networks weights
   * either per pattern, or per epoch
   */
  private final void updateWeights()
  {
    for (int j=0; j < outputLayer.length; j++)  outputLayer[j].realizeBackwardLinks();
    for (int j=0; j < hiddenLayer.length; j++)  hiddenLayer[j].realizeBackwardLinks();
  }
//---------------------------------------------------------------------//
 private double testThisPattern(Instance p)
 {
   runThisInput(p);
      
   double totalError = 0.0, value, diff;
   double[] correctValue = getCorrectOutput(p.getClassID()); 
   
   for (int i=0; i < outputLayer.length; i++)
   {
     value = outputLayer[i].nodeFunction();

     diff = Math.abs((value - correctValue[i]));
     totalError += diff;
   }
   
   return totalError;
 }
//---------------------------------------------------------------------//
 private void printResultsX(long secs, Vector stats, Vector<Instance> patterns)
 {
   maxMappingError = 0.05;
   
   ///////////////////////////////////////////////////////////
   ///// run through all patterns again,  with no training
   int sux = 0;
   double plainError, patternError;
   Instance instance;

   gui.clearOutput();
   gui.print("FAILED PATTERN CLASSIFICATIONS:");
   
   for (int i=0; i < patterns.size(); i++)
   {
     instance     = patterns.elementAt(i);      
     patternError = testThisPattern(instance);
     
     if (patternError < (maxMappingError * outputLayer.length))
     {
       sux++;
     }
     else
     {
       gui.print("\n" + instance.getLabel() + " failed, value = " + patternError + " (avg = " + (patternError/outputLayer.length) + ")");
     }
   }
   
   gui.print("\n------------------------------------");
   
   if (stats != null) 
   {
     plainError = ((Double)stats.lastElement()).doubleValue();
     gui.print("\n\nMSE for last epoch : " + plainError);
     gui.print("\nEpochs run         : " + epochs);
     gui.print("\nTime elapsed       : " + secs + " secs");
   }    
   double pct = ((double)sux / (double)patterns.size()) * 100.0;
   gui.print("\nACCURACY : " + pct + "%");
 }
//--------------------------------------------------------------------//
 public Vector<Instance> getTrainingData()
 {
   return this.patterns_train;
 }
//--------------------------------------------------------------------//
 public Vector<Instance> getTestData()
 {
   return this.patterns_test;
 }
//--------------------------------------------------------------------//
 public int getNumberOfTrainingSamples()
 {
   return patterns_train.size();
 }
//-------------------------------------------------------------------//
 public int getNumberOfTestingSamples()
 {
   return patterns_test.size();
 }
//---------------------------------------------------------------------//
 /**
  * In case of multiple hiddenlayers, the last one is returned!
  */
 public Neuron[] getHiddenLayer()
 { 
   return layerVector.elementAt(layerVector.size()-2);
 }
//---------------------------------------------------------------------//
 public Neuron[] getOutputLayer()
 { 
   return outputLayer;
 }
//---------------------------------------------------------------------//
 public double[] getLayerActivation(Neuron[] layer)
 {
   double[] actualOutput = new double[layer.length];
   
   for (int i=0; i < actualOutput.length; i++)
   {
     actualOutput[i] = layer[i].nodeFunction();
   }
   return actualOutput;   
 }
//---------------------------------------------------------------------//
 public void startTraining(double lr, double mom, int epo, boolean adaptiveLR)
 {
   annCfg.setLearningR(lr);
   annCfg.setMomentum(mom);
   annCfg.setEpochs(epo);
   annCfg.setUseAdaptiveLearningRate(adaptiveLR);
   
   startTraining();
 }
//---------------------------------------------------------------------//
 public void startTraining()
 {
   // LR, momentum, epochs...
   loadSimpleParameters();
   
   Thread t = new Thread(this);
   t.setPriority(Thread.NORM_PRIORITY);
   t.start();
 }
//--------------------------------------------------------------------//   
 public void startTesting()
 {
   printResultsX(0, null, patterns_test);
 }   
//---------------------------------------------------------------------//   
 public void run()
 {
   startTrainingX();
   gui.notify_finished();
 }
//---------------------------------------------------------------------//  
}