package sharkey_mod_nlp.dataset;

import java.io.Serializable;
import java.util.*;
import sharkey_mod_nlp.treemodel.*;

/**
 * Dataset hierarchy:
 * 
 *    Dataset--
 * 	          |
 *            -- Instance--
 *                        |
 *                        -- Attribute
 */
public final class Dataset implements Serializable
{
  private String[] attributeNames, classNames;
  private String datasetName;
  
  private ValuesForAttribute[] valuesForAttribute;
  private Vector<Instance> dataset;
  private double maxValue_real, minValue_real;
  
  private int numberOfClasses;
  private long globalMaxAttributeValue, maxValue_int, minValue_int;
  
  private Lexicon lexicon;
  private Vector<Instance> dsTrain, dsVal, dsTest;
  private Vector<ParseNode> parseTrees; // for RAAM-training
  
  public Dataset(String name, int nClasses)
  {
    datasetName        = name;
    numberOfClasses    = nClasses;
    dataset            = new Vector<Instance>();
    parseTrees         = new Vector<ParseNode>();
    lexicon            = new Lexicon();
  }
//---------------------------------------------------------------------//
  public String getName()
  {
    return this.datasetName;
  }
//---------------------------------------------------------------------//
  public Lexicon getLexicon()
  {
    return lexicon;
  }
//---------------------------------------------------------------------//
  public static double[] getExpandedPatternFromInt(int setBitNo, int patternWidth)
  {
    double[] bp = new double[patternWidth];
    
    for (int i=0; i < patternWidth; i++) bp[i] = 0.0;
    bp[setBitNo] = 1.0;
    
    return bp;
  }
//---------------------------------------------------------------------//
  public final void addLexiconEntry(String[] data, int id, int patternWidth)
  {
    //  1st attribute is word (character squence), the remaining ones
    // are a list of possible POS classes; usually just one.
    
    int pos_mask = 0;
    int binPow;
    int[] bits = new int[patternWidth];
    
    // Create bitmask by OR'ing with orthogonal bitmasks corresponding to each POS id
    for (int i=1; i < data.length; i++)
    {
      binPow = (int) Math.pow(2, lexicon.getPosIDFor(data[i]));
      pos_mask |= binPow;
      
      int posID = lexicon.getPosIDFor(data[i]);
      
      if (posID != -1)
      {
        // System.out.println("NO POS FOR " + data[i]);
        bits[posID] = 1;
      }      
    }
    
    lexicon.addEntry(data[0], pos_mask, bits, id, patternWidth);
  }
//---------------------------------------------------------------------//
  public final void addPOS(String[] data, int id)
  {
    lexicon.addPOS_TAG(data, id);
  }
//---------------------------------------------------------------------//
  public final void addParseStructure(ParseNode root)
  {
    parseTrees.addElement(root);
  }
//---------------------------------------------------------------------//
  public Vector<ParseNode> getParseTrees()
  {
    return this.parseTrees;
  }
//---------------------------------------------------------------------//
  public void compile(int bitpattternWidth)
  {
    // Must be done FIRST!
    for (int i=0; i < dataset.size(); i++)
    {
      dataset.elementAt(i).createAttributeValues_NLP(lexicon);
    }
    
    findMinMaxValues();
    findIntegerMinMaxValues();
    
    String maxBinStr = Integer.toBinaryString(lexicon.getWordsInDictionary().size());
    int nBits = maxBinStr.length();
    
    // System.out.println("compactPattern width = " + nBits);
    // System.out.println("words in dic         = " + lexicon.getWordsInDictionary().size());
    // System.out.println("maxBinStr            = " + maxBinStr);
    
    for (int i=0; i < dataset.size(); i++)
    {
      dataset.elementAt(i).create_plain_ANN_format(maxValue_real, minValue_real);
      dataset.elementAt(i).createExpandedPattern(bitpattternWidth);
      
      // this creates a 'no info' pattern for each word; used by the SRN encoder
      // which should NOT be given any structural/grammatical information.
      dataset.elementAt(i).createSimpleExpandedPattern(lexicon, nBits, true);
    }
  }
//---------------------------------------------------------------------//
  /**
   * Sets (and replaces) the existing set of instances
   * already stored here, if any.
   * 
   * !PS!: This method may NOT be compatible with the existing
   * data-analyzing and preparation code! The instances supplied
   * here are supposed to be self-contained with regard to any (meta-)info
   * attatched to instances when loaded the normal way. 
   * (=> as when loading a dataset from file). 
   */
  public void setInstances(Vector<Instance> newInstances)
  {
    dataset = newInstances;
  }
//---------------------------------------------------------------------//
  /**
   * tranforms the raw version of the data in each instance in the dataset; makes
   * them into structured objects of type Instance
   */
  public final void addInstance(String[] attributes, int classID, String instLabel)
  {
    Attribute[] attribs = new Attribute[attributes.length];
    double value;
    
    for (int i=0; i < attribs.length; i++)
    {
      try
      {
        value      = Double.parseDouble(attributes[i]);
        attribs[i] = new Attribute(i, "Attribute #" + i, value);
      }
      catch (NumberFormatException e)
      {
        // set the attribute-value as text
        attribs[i] = new Attribute(i, "Attribute #" + i, attributes[i]);
      }
    }
    
    dataset.addElement(new Instance(instLabel, attribs, classID));
  }
//---------------------------------------------------------------------//  
  public final String[] getAttributeNames()
  {
    return attributeNames;
  }
//---------------------------------------------------------------------//
  private void sortArray_ascending(long[] values)
  {
  	long tmp;
  	
  	for (int i=0; i < values.length; i++)
  	{
  	  for (int j=i+1; j < values.length; j++)
  	  {
  		if (values[i] > values[j])
  		{
  		  tmp       = values[i];
  		  values[i] = values[j];
  		  values[j] = tmp;
  		}
  	  }
    }
  }
//---------------------------------------------------------------------//
  public double getMulWeightsForClasses(Vector<Instance> instances, int classID)
  {
  	double cumWt = 1.0;
  	Instance inst;
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  		inst = instances.elementAt(i);
  		
 			if (inst.getClassID() == classID)
 			{
 				cumWt *= inst.getWeight();
 			}
  	}
  	return cumWt;
  }
//---------------------------------------------------------------------//
  public double getCumulativeWeightsForClasses(Vector<Instance> instances, int classID)
  {
  	double cumWt = 0.0;
  	Instance inst;
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  		inst = instances.elementAt(i);
  		
 			if (inst.getClassID() == classID)
 			{
 				cumWt += inst.getWeight();
 			}
  	}
  	return cumWt;
  }
//---------------------------------------------------------------------//
  /**
   * Used in computing entropy and getting class-distribution for a 
   * set of data; usually the the set of instances reserved for
   * training_data, and subsets thereof...
   */
  public double getFractionBelongingToClass(Vector<Instance> instances, int classID)
  {
  	int cCount = 0;
  	Instance inst;
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  		inst = instances.elementAt(i);
  		
 			if (inst.getClassID() == classID) cCount++;  		
  	}  	
  	return (double)cCount / (double)instances.size();
  }
//---------------------------------------------------------------------//
  /**
   * returns a list of all distinct real-values found for attribute 'aNo' 
   * Used by 'simplifyRealValues()' in class Instance.
   * If the param 'instances' is null, the whole dataset will be used.
   */
  public Vector<Double> getValuesForAttributeIn(Vector<Instance> instances, int aNo)
  {
  	if (instances == null) instances = dataset;
  	
  	double tmpValue;
  	Vector<Double> v = new Vector<Double>();
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  		tmpValue = instances.elementAt(i).getAttribute(aNo).getValue();
  
  		if (!containsValue(v, tmpValue))
  		{
  			v.addElement(new Double(tmpValue));
  		}  		
  	}     	
  	return v;
  }
//---------------------------------------------------------------------//
  /**
   * returns a list of all distinct simple-values found for attribute 'aNo' 
   * -> Currently used by the DecisionTree-classifier only...
   * If the param 'instances' is null, the whole dataset will be used.
   */
  public Vector<Integer> getSimpleValuesForAttributeIn(Vector<Instance> instances, int aNo)
  {
  	if (instances == null) instances = dataset;
  	
  	long tmpValue;
  	Vector<Double>  v  = new Vector<Double>();
  	Vector<Integer> vI = new Vector<Integer>();
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  		tmpValue = instances.elementAt(i).getAttribute(aNo).getSimpleValue();
  
  		if (!containsValue(v, tmpValue))
  		{
  			v.addElement(new Double(tmpValue));
  			vI.addElement(new Integer((int)tmpValue));
  		}  		
  	}     	
  	return vI;
  }
//---------------------------------------------------------------------//
  public double[] getRealValuesForAttributeIn(Vector<Instance> instances, int aNo)
  {
  	if (instances == null) instances = dataset;

  	// optimization for attribute transformations (simplify & reduce).
    // These values are only set when 'instances' are the whole dataset
  	if (valuesForAttribute[aNo].getRealValues() != null)
  	{
  		return valuesForAttribute[aNo].getRealValues(); 
  	}
  	//-------- already computed and filled in (set = whole dataset) -------
  	
  	double tmpValue;
  	Vector<Double>  v  = new Vector<Double>();
  	Vector<Integer> vI = new Vector<Integer>();
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  		tmpValue = instances.elementAt(i).getAttribute(aNo).getValue();
  
  		if (!containsValue(v, tmpValue))
  		{
  			v.addElement(new Double(tmpValue));
  			vI.addElement(new Integer((int)tmpValue));
  		}  		
  	}   
  	
  	double[] rvA = new double[v.size()];
  	for (int i=0; i < rvA.length; i++)
  	{
  		rvA[i] = v.elementAt(i).doubleValue();
  	}
  	
  	// Only when 'instances' are the whole dataset;
  	// we can't keep track of every possible subset...
  	if (instances.equals(dataset))
  	{
  		valuesForAttribute[aNo].setRealValues(rvA);
  	}
  	
  	return rvA;
  }
//---------------------------------------------------------------------//
  /**
   * Used by the DTC in the entropy calculations; a set of instances
   * where all are of the same class implies ZERO entropy...
   */
  public Vector<Integer> getDistinctClassesIn(Vector<Instance> instances)
  {
  	if (instances == null) instances = dataset;
  	
  	int tmpClass;
  	Vector<Double> v   = new Vector<Double>();
  	Vector<Integer> vI = new Vector<Integer>();
  	
  	for (int i=0; i < instances.size(); i++)
  	{
  	  tmpClass = instances.elementAt(i).getClassID();
  		
  	  if (!containsValue(v, tmpClass))
  	  {
  		v.addElement(new Double(tmpClass));
  		vI.addElement(new Integer(tmpClass));
  	  }
  	}
  	
  	return vI;
  }
//---------------------------------------------------------------------//  
  private final boolean containsValue(Vector<Double> v, double value)
  {
  	for (int i=0; i < v.size(); i++)
  	{
  	  if (v.elementAt(i).doubleValue() == value) return true;
  	}  	
  	return false;
  }
//---------------------------------------------------------------------//
  /**
   * This is used to facilitate the task of dynamically/automatically
   * suggest the new number of different values when downsizing / simplyfying
   * the data for use in the NaiveBayes & DecisionTree-Classifier. And whether
   * it is neccessarry to do so...
   */
  private void XcomputeSimpleAttributeDiversity()
  {
  	Instance inst;
    Attribute[] attrib;
    Vector<Double>[] diffValueVector = new Vector[getNumberOfAttributes()];
    // simpleAttributeDiversityCount    = new int[getNumberOfAttributes()];
    int value;
    
    for (int i=0; i < dataset.size(); i++)
    {
      inst   = dataset.elementAt(i);
      attrib = inst.getAttributes();
      
      for (int j=0; j < attrib.length; j++)
      {
      	value = (int)attrib[j].getSimpleValue();
        if (diffValueVector[j] == null) diffValueVector[j] = new Vector<Double>();
        
        if (isNewValue(diffValueVector[j], 0.0, value))
        {
          diffValueVector[j].add(new Double(value));
          // simpleAttributeDiversityCount[j]++;
        }
      }
    }
  }
//---------------------------------------------------------------------//  
  /**
   * Finds the number of different values found in the dataset 
   * for each attribute. Values 'less different' than given by the
   * parameter 'minDiff' are considered equal in this respect.
   */
  private void XcomputeAttributeDiversity(double minDiff)
  {
    Instance inst;
    Attribute[] attrib;
    Vector<Double>[] diffValueVector = new Vector[getNumberOfAttributes()];
    // attributeDiversityCount          = new int[getNumberOfAttributes()];
    
    for (int i=0; i < dataset.size(); i++)
    {
      inst   = dataset.elementAt(i);
      attrib = inst.getAttributes();
      
      for (int j=0; j < attrib.length; j++)
      {
        if (diffValueVector[j] == null) diffValueVector[j] = new Vector<Double>();
        
        if (isNewValue(diffValueVector[j], minDiff, attrib[j].getValue()))
        {
          diffValueVector[j].add(new Double(attrib[j].getValue()));
          // attributeDiversityCount[j]++;
        }
      }
    }
  }
//---------------------------------------------------------------------//  
  private boolean isNewValue(Vector<Double> v, double minDiff, double value)
  {
    double diff;
    
    if (v.isEmpty()) return true;
    
    for (int i=0; i < v.size(); i++)
    {
      diff = Math.abs(value - v.elementAt(i).doubleValue());      
      if (diff <= minDiff) return false;
    }    
    return true;
  }
//---------------------------------------------------------------------//  
  /**
   * returns the data as a set of instances; a list of class Instance objects
   */
  public final Vector<Instance> getInstanceData()
  {
    return dataset;
  }
//---------------------------------------------------------------------//
  /**
   * returns a 'list-of-lists' with instance data grouped by their classifications;
   * used for picking out a 'reasonably representative' set of training data...
   */
  private final Vector<Instance>[] getDataGroupedByClass()
  {
    Vector<Instance>[] v = new Vector[numberOfClasses];    
    Instance inst;
    
    for (int i=0; i < v.length; i++) v[i] = new Vector<Instance>(); 
    
    for (int i=0; i < dataset.size(); i++)
    {
      inst = dataset.elementAt(i);
      v[inst.getClassID()].addElement(inst);
    }    
    return v;
  }
//---------------------------------------------------------------------//
  /**
   * This method is supposed to be called with data from ANOTHER
   * instance of 'Dataset', otherwise the returned set will be
   * EXACTLY the same as the param 'srcData'!
   * 
   * Used by the ADABOOSTER to get separate copies of the same
   * instances. I.e. 3 different instantiations of the Dataset-
   * class will have the same data-instances, but as individual
   * copies (not just pointers to the same object)...
   */
  public Vector<Instance> getInstancesWithSameID(Vector<Instance> srcData)
  {
  	Vector<Instance> sameSet = new Vector<Instance>();
  	int id;
  	
  	for (int i=0; i < srcData.size(); i++)
  	{
  		id = srcData.elementAt(i).getInstanceNumber();
  		sameSet.addElement(getInstanceNo(id));
  	}
  	return sameSet;
  }
//---------------------------------------------------------------------//
  private Instance getInstanceNo(int idNo)
  {
  	for (int i=0; i < dataset.size(); i++)
  	{
  		if (dataset.elementAt(i).getInstanceNumber() == idNo)
  		{
  			return dataset.elementAt(i);
  		}
  	}
  	
  	System.out.println("ERROR : Could not find instance with ID " + idNo);
  	return null; // ehee, nooooo
  }
//---------------------------------------------------------------------//
  /**
   * Testing-method ensuring same sets of instances
   * when developing the adabooster... (for each instantiation of it)
   */
  public void splitDatasetAndStorePartitions(double pctTrain, double pctVal)
  {
  	dsTrain = new Vector<Instance>();
  	dsVal   = new Vector<Instance>();
  	dsTest  = new Vector<Instance>();
  	
  	splitDataset(pctTrain, pctVal, dsTrain, dsVal, dsTest);
  }
//--------------------------------------------------------------------------------//
  public void getPartitionedDatasets(Vector<Instance> data_train, Vector<Instance> data_validation, Vector<Instance> data_test)
  {
  	for (int i=0; i < dsTrain.size(); i++) data_train.addElement(dsTrain.elementAt(i));
  	for (int i=0; i < dsVal.size(); i++)   data_validation.addElement(dsVal.elementAt(i));
  	for (int i=0; i < dsTest.size(); i++) data_test.addElement(dsTest.elementAt(i));
  }
//--------------------------------------------------------------------------------//
  /**
   * While the selection of training-samples is randomized to ensure fair distribution(?), we
   * also SHOULD ensure that this will cover some instances from all CLASSES.
   *   - (I.e. the general assumption that a given dataset for training must be 'reasonably'
   * representative of the corresponding (un)limited set of real data, out there...)
   * 
   * NB: The rough method here will cause nearly all samples of some classes to be included 
   * in the training set..., if a class has very few samples relative to the total amount of
   * training samples.
   */
  public void splitDataset(double pctForTraining, double pctForValidation, Vector<Instance> data_train, Vector<Instance> data_validation, Vector<Instance> data_test)
  {
    data_train.removeAllElements();
    data_test.removeAllElements();
    if (data_validation != null) data_validation.removeAllElements();
    
    int nTrainElems      = (int) (pctForTraining   * getNumberOfInstances());
    int nValidationElems = (int) (pctForValidation * nTrainElems);
    int tmpElem;
    
    // if dataset too small, use ALL samples for training... (probably just a dummy-dataset)
    if (pctForTraining >= 1.0 || getNumberOfInstances() < 25 || nValidationElems <= 1)
    {
      pctForTraining   = 1.0;
      pctForValidation = 0.0;
      nTrainElems      = getNumberOfInstances();
    }
    
    // these are *copies* of references, so we may safely (non-destructively) remove elements
    // from these vectors, in vArr[], causing no change to the original dataset... :)
    Vector<Instance>[] vArr = getDataGroupedByClass();
    boolean progress = true;
    
    while (data_train.size() < nTrainElems && progress)
    {
      progress = false;
      
      for (int i=0; i < vArr.length; i++)
      {
        // don't spend it all on training; save at least 2 for testing
        if (vArr[i].size() > 2)
        {
          tmpElem = (int) (Math.random() * vArr[i].size());
          data_train.addElement(vArr[i].remove(tmpElem));
          progress = true;
          if (data_train.size() >= nTrainElems) break;
        }        
      }
    }
    
    // just add the any samples to satisy the size of training-set...
    while (data_train.size() < nTrainElems)
    {
      for (int i=0; i < vArr.length; i++)
      {
        // don't spend it all on training; save at least 2 for testing
        if (!vArr[i].isEmpty())
        {
          tmpElem = (int) (Math.random() * vArr[i].size());
          data_train.addElement(vArr[i].remove(tmpElem));
          if (data_train.size() >= nTrainElems) break;
        }        
      }
    }
    
    ////////////////////////////////////////////////////////////
    ///// Use the given ratio to get samples to use for the  
    ///// validation-set: these are taken from the TRAINING DATA!
    if (nValidationElems > 0 && data_validation != null)
    {
      for (int i=0; i < nValidationElems; i++)
      {
        data_validation.addElement(data_train.remove(0));
      }
    }
    
    ////////////////////////////////////////////////////////
    ///// Now put the remaining samples into the 'test' bag
    boolean empty = false;
    
    while (!empty)
    {
      empty = true;
      
      for (int i=0; i < vArr.length; i++)
      {
        if (!vArr[i].isEmpty()) 
        {
          empty = false;          
          for (int j=0; j < vArr[i].size(); j++)
          {
            data_test.addElement(vArr[i].elementAt(j));
          }
          vArr[i].removeAllElements();
        }        
      }
    }
  }
//---------------------------------------------------------------------//
  public void XresetAttributeStatistics()
  {
    /*
    valuesForAttribute = new ValuesForAttribute[numberOfAttributes];
    
    for (int i=0; i < numberOfAttributes; i++)
    {
      valuesForAttribute[i] = new ValuesForAttribute();
    }
    */
  }
//---------------------------------------------------------------------//
  /**
   * Creates a simplified value for the specified attribute, for all 
   * instances in our dataset.
   */
  public final void XsimplifyValues(int atrNo, int nValues, boolean makeIntegerValues)
  {
    /*
    int origDiversity = getAttributeDiversityFor(atrNo);

    if (origDiversity < nValues)  nValues = origDiversity; 
    
    Instance inst;
    
    for (int i=0; i < dataset.size(); i++)
    {
      inst = dataset.elementAt(i);
      
      if (makeIntegerValues)
      {
      	inst.simplifyIntegerValues(this, nValues, atrNo);
      }
      else // here we are to reduce the diversity of the raw values (in double/real format)
      {
      	inst.simplifyRealValues(this, nValues, atrNo);
      }
    }
    */
  }
//---------------------------------------------------------------------//  
  private void findIntegerMinMaxValues()
  {
    maxValue_int = -999999999;
    minValue_int =  999999999;
    
    Attribute[] attribs;
    
    for (int i=0; i < dataset.size(); i++)
    {
      attribs = dataset.elementAt(i).getAttributes();
      
      for (int j=0; j < attribs.length; j++)
      {
        if (attribs[j].getSimpleValue() > maxValue_int) maxValue_int = attribs[j].getSimpleValue(); 
        if (attribs[j].getSimpleValue() < minValue_int) minValue_int = attribs[j].getSimpleValue(); 
      }
    }  
  }
//---------------------------------------------------------------------//
  private void findMinMaxValues()
  {
  	maxValue_real = -999999999;
  	minValue_real =  999999999;
  	
  	Attribute[] attribs;
  	
  	for (int i=0; i < dataset.size(); i++)
  	{
      attribs = dataset.elementAt(i).getAttributes();
  		
      for (int j=0; j < attribs.length; j++)
      {
  		  if (attribs[j].getValue() > maxValue_real) maxValue_real = attribs[j].getValue(); 
  		  if (attribs[j].getValue() < minValue_real) minValue_real = attribs[j].getValue(); 
  	  }
  	}
  }
//---------------------------------------------------------------------//
  /**
   * Maximum INTEGER value for the whole dataset!
   * Used in some classifiers to create a hash-array for storing attribute-value
   * counts, with values as index: naturally this refers to INTEGER values, 
   * and so this max-value is gleaned from the attribute.getSimpleValue()...
   */
  public final long getGlobalMaxAttributeValue()
  {
    return globalMaxAttributeValue;
  }
//---------------------------------------------------------------------//  
  public final void setAttributeNames(String[] attribNames)
  {
    this.attributeNames = attribNames;
  }
//---------------------------------------------------------------------//  
  public final void setClassNames(String[] cNames)
  {
    this.classNames = cNames;
  }
//---------------------------------------------------------------------//
  public String getClassName(int classID)
  {
    return classNames[classID];
  }
//---------------------------------------------------------------------//
  public double[] reduceDiversityReal(double[] values, int nValues, int atrNo)
  {
  	double[] redRng = valuesForAttribute[atrNo].getReducedRangeR();
		if (redRng != null) return redRng;
    //-------- already computed and filled in ---------------
		
  	double[] reducedRange = new double[nValues];
  	Vector<Double> rv = new Vector<Double>();
  	
  	int inc = (int)((double)values.length / (double)nValues);
  	int idx = 0, k = 0;
  	
  	while ((rv.size() < nValues) && (k < values.length))
  	{
  		rv.addElement(new Double(values[k]));
  		
  		k += inc;
  		idx++;
  	}

 		int xtra = nValues - rv.size();
 		
  	for (int i=0; i < xtra; i++)  rv.addElement(new Double(values[idx++]));
  	
  	for (int i=0; i < rv.size(); i++) 
  	{
  		reducedRange[i] = rv.elementAt(i).doubleValue();
  	}
  	
  	valuesForAttribute[atrNo].setReducedRangeR(reducedRange);  	
  	return reducedRange;
  }
//---------------------------------------------------------------------//
  public long[] reduceDiversityInteger(long[] values, int nValues, int atrNo)
  {
  	long[] redRng = valuesForAttribute[atrNo].getReducedRangeI();
		if (redRng != null) return redRng;
    //-------- already computed and filled in ---------------
		
  	long[] reducedRange = new long[nValues];
  	Vector<Long> rv = new Vector<Long>();
  	
  	int inc = (values.length / nValues);
  	int idx = 0, k = 0;
  	
  	while (rv.size() < nValues)
  	{
  		rv.addElement(new Long(values[k]));
  		
  		k += inc;
  		idx++;
  	}

 		int xtra = nValues - rv.size();
  	for (int i=0; i < xtra; i++)  rv.addElement(new Long(values[idx++]));
  	
  	for (int i=0; i < rv.size(); i++) 
  	{
  		reducedRange[i] = rv.elementAt(i).longValue();
  	}
  	
  	valuesForAttribute[atrNo].setReducedRangeI(reducedRange);  	
  	return reducedRange;
  }
//---------------------------------------------------------------------//
  /**
   * creates a set of integer values from a set of double values, preserving
   * the internal relative sizes and distribution. I think...
   */
  public long[] createIntegerValuesForAttribute(int atrNo)
  {
  	if (valuesForAttribute[atrNo].getIntegerValues() != null)
  	{
  		return valuesForAttribute[atrNo].getIntegerValues();
  	}
  	//-------- already computed and filled in -------------
  	
    Vector<Double> v = getValuesForAttributeIn(dataset, atrNo);    
    int maxDecimals  = getMaxDecimalsPresentIn(v);
    long[] allValues = new long[v.size()];
    
    for (int i=0; i < v.size(); i++)
    {
    	allValues[i] = getIntegerValue(v.elementAt(i).doubleValue(), maxDecimals);
    }

    sortArray_ascending(allValues);
    long smallest = allValues[0];
    
    // substract the smallest value from all values...
    for (int i=0; i < allValues.length; i++) allValues[i] -= smallest;
    
    valuesForAttribute[atrNo].setIntegerValues(allValues);
    valuesForAttribute[atrNo].setSmallestValueI(smallest);
    valuesForAttribute[atrNo].setMaxDecimals(maxDecimals);

    return allValues;
  }
//------------------------------------------------------------------------------//  
  private int getMaxDecimalsPresentIn(Vector<Double> v)
  {
  	String decStr = "";  	
  	int maxD = 0;
  	
  	for (int i=0; i < v.size(); i++)
    {
  		decStr = v.elementAt(i).doubleValue() + "";
  		
  		if (decStr.indexOf(".") != -1)
    	{
  			// all digits AFTER the "."
    		decStr = decStr.substring(decStr.indexOf(".")+1);
    	}
  		else decStr = "";
  		
  		if (decStr.length() > maxD) maxD = decStr.length();
    }  	
  	return maxD;
  }
//---------------------------------------------------------------------//
  private long getIntegerValue(double value, int maxDecimals)
  {
  	String sVal = getPaddedValueStrWithoutComma(value, maxDecimals);  	
  	return Long.parseLong(sVal);
  }
//---------------------------------------------------------------------//
  /**
   * Takes in a double, like 1.503432 and converts it to the string  '1503432'
   * To handle the case where a larger value, '302.51' and a smaller 1.73012'
   * would become:  30251 (large)
   *               173012 (small)
   *              
   * we use the 'maxDecimals' to ensure all get the same number of decimals,
   * by padding a number of 0's to the shorter string: this keeps the relative
   * magnitudes consistent. 
   */
  private String getPaddedValueStrWithoutComma(double value, int maxDecimals)
  {
  	String valStr = value + "";
    String sVal   = null;
    int nDecimals = 0;
    
    if (valStr.indexOf(".") != -1)
    {
    	sVal = valStr.substring(0, valStr.indexOf("."));
    	sVal += valStr.substring(valStr.indexOf(".")+1);
    	
    	nDecimals = valStr.substring(valStr.indexOf(".")+1).length();
    }
    else sVal = value + "";
    
    int zerosToAdd = (maxDecimals - nDecimals);
    
    for (int i=0; i < zerosToAdd; i++) sVal += "0";
    
    return sVal;
  }
//---------------------------------------------------------------------//
  /**
   * returns the size of the dataset; the number of instances
   */
  public int getNumberOfInstances() 
  {
    return dataset.size(); 
  }
//---------------------------------------------------------------------//
  public int getNumberOfClasses() 
  { 
    return numberOfClasses; 
  }
//---------------------------------------------------------------------//
  public int getNumberOfAttributes() 
  { 
    return -1; 
  }
//---------------------------------------------------------------------//
  public Dataset cloneDataset()
  {
    // 1. clone all instances
    Vector<Instance> clonedInstances = new Vector<Instance>();
    for (int i=0; i < dataset.size(); i++)
    {
        clonedInstances.addElement(dataset.elementAt(i).cloneMe());
    }
    Dataset ds = new Dataset(datasetName, numberOfClasses);
    ds.dataset = clonedInstances;
    
    // no point to clone these datastructures...    
    ds.attributeNames     = attributeNames;
    ds.classNames         = classNames;
    ds.valuesForAttribute = valuesForAttribute;
    ds.maxValue_real     = maxValue_real;
    ds.minValue_real     = minValue_real;
    ds.globalMaxAttributeValue = globalMaxAttributeValue;
    
    return ds;
  }
//---------------------------------------------------------------------// 
}

/**
 * 
 * Helper class for storing values, ranges, tranformations...
 * that has to be recalculated after changes to the dataset; most
 * likely a reduce-and-simplify operation. Optimization to avoid
 * redundant operations... 
 *
 */
final class ValuesForAttribute
{
	private long[] reducedRangeI;
	private long[] integerValues;
	
	private double[] realValues;
	private double[] reducedRangeR;
	
	private long smallestValueI;
	private double smallestValueR;
	
	private int maxDecimals;
//---------------------------------------------------------------------//
	public ValuesForAttribute()
	{
		smallestValueI = Long.MAX_VALUE;
		maxDecimals    = -1;
	}
//-------------------------------------------------------------------------//
	public int    getMaxDecimals()     { return maxDecimals;    }
	public long   getSmallestValueI()  { return smallestValueI; }
	public double getSmallestValueR()  { return smallestValueR; }
	
	public long[] getReducedRangeI()   { return reducedRangeI; }
	public long[] getIntegerValues()   { return integerValues; }
	
	public double[] getReducedRangeR() { return reducedRangeR; }
	public double[] getRealValues()    { return realValues;    }
//-------------------------------------------------------------------------//
	public void setReducedRangeI(long[] reducedRng)
	{
		reducedRangeI = reducedRng;
	}
//---------------------------------------------------------------------//
	public void setReducedRangeR(double[] reducedRng)
	{
		reducedRangeR = reducedRng;
	}
//---------------------------------------------------------------------//
	public void setIntegerValues(long[] values)
	{
		integerValues = values;
	}
//---------------------------------------------------------------------//
	public void setRealValues(double[] values)
	{
		realValues = values;
	}
//---------------------------------------------------------------------//
	public void setSmallestValueI(long minV)
	{
		smallestValueI = minV;
	}
//---------------------------------------------------------------------//
	public void setSmallestValueRI(double minV)
	{
		smallestValueR = minV;
	}
//---------------------------------------------------------------------//
	public void setMaxDecimals(int n)
	{
		maxDecimals = n;
	}
//---------------------------------------------------------------------//
}