package sharkey_mod_nlp.dataset;

import java.io.Serializable;

//-===============================================================--//
/**
 * This is the container class for a data-sample (instance) in the dataset;
 * All attributes are stored here.
 * 
 * ~ An instance is the equvivalent of the ANN's Pattern-class; yielding 1 inputNode
 *   pr. attribute, or x nodes if we use expanded patterns: then each attribute has
 *   a n-width bitpattern. Useful for encoding more info pr. attrib when they
 *   are natural language "words". (lexemes)
 */
public final class Instance implements Serializable
{
  private static int INSTANCE_COUNTER = 0x00;
	
  private String label;
  private Attribute[] attributes;
  private int classID, id;
  
  private ExpandedPattern[] expandedPattern_POS_INFO;
  private ExpandedPattern[] expandedPattern_NO_INFO;
  private double[] annData;
  private double weight, minWeight, maxWeight;
  
  /**
   * Generation of NeuralNet values:
   * Values of all attributes are normalized into the range [0.0, 1.0].
   */  
  public Instance(String label, Attribute[] attribs, int classID)
  {
    this.attributes = attribs;
    this.classID    = classID;
    this.label      = label;
    this.id         = INSTANCE_COUNTER++;
    
    weight     = 1.0;
    minWeight  = 1.0;
    maxWeight  = 1.0;
  }
//---------------------------------------------------------------------//
  /**
   * For cloning...
   */
  public Instance(String label, int id, int classID)
  {
  	this.label   = label;
  	this.id      = id;
  	this.classID = classID;
  }
//---------------------------------------------------------------------//
  public Instance cloneMe()
  {
  	Instance cInst = new Instance(label, id, classID);
    
  	cInst.setWeight(weight);
  	cInst.setWeightBounds(minWeight, maxWeight);
  	
  	Attribute[] cAttribs = new Attribute[attributes.length];
  	
  	for (int i=0; i < attributes.length; i++)
  	{
  		cAttribs[i] = attributes[i].cloneMe();
  	}
  	
  	cInst.attributes = cAttribs; 
  	return cInst;
  }
//---------------------------------------------------------------------//
  public int getInstanceNumber()
  {
  	return id;
  }
//---------------------------------------------------------------------//
  public static void resetInstanceCounter()
  {
  	Instance.INSTANCE_COUNTER = 0x00;
  }
//---------------------------------------------------------------------//
  public double getUpperWeightBound()
  {
  	return maxWeight;
  }
//---------------------------------------------------------------------//
  public double getLowerWeightBound()
  {
  	return minWeight;
  }
//---------------------------------------------------------------------//
  public void setWeightBounds(double min, double max)
  {
  	minWeight = min;
  	maxWeight = max;
  }
//---------------------------------------------------------------------//
  public boolean changeWeight(double dWt)
  {
  	weight += dWt;
  	
    if (weight < minWeight)
    {
    	weight = minWeight;
    	return false;
    }
  	else if (weight > maxWeight)
  	{
  		weight = maxWeight;
  		return false;
  	}
  	return true;
  }
//---------------------------------------------------------------------//
  public void setWeight(double wt)
  {
  	weight = wt;
  	
  	     if (weight < minWeight) weight = minWeight;  
  	else if (weight > maxWeight) weight = maxWeight;
  }
//---------------------------------------------------------------------//
  public double getWeight()
  {
  	return weight;
  }
//---------------------------------------------------------------------//
  /**
   * This will reduce the number of different values an attribute has, to 'nValues'.
   * Typically, we will want to transform a real value range of, say, {0.01, 0.87} to
   * values like Low, Medium, High... {0, 1, 2}. Also, an attribute that takes on quite many
   * different INTEGER values, like 0,2,3,4,5,6,9,14,43,60.. can be useful to shrink to
   * a more manageable range of {0, 6}.
   * 
   * Instead of simply creating a linear range of integers from 0 to nValues-1, we
   * try to retain some aspect of relative distances between values. This seems to
   * make sense when considering classifiers like Nearest Neighbor and operations
   * like cluster analysis. 
   */
  public void simplifyIntegerValues(Dataset ds, int nValues_new, int atrNo)
  {
  	double[] rawValues = ds.getRealValuesForAttributeIn(null, atrNo);
  	double origValue   = attributes[atrNo].getValue();
  	
  	int oldIndex = -1;
  	for (int i=0; i < rawValues.length; i++)
  	{
  		if (origValue == rawValues[i])
  		{
  			oldIndex = i+1; // to avoid having to write "rawValues.length-1" 
  			break;
  		}
  	}
  	
    long[] allValues = ds.createIntegerValuesForAttribute(atrNo);
    long[] newValues = ds.reduceDiversityInteger(allValues, nValues_new, atrNo);    

    double oldPctOfMaxIdx = ((double)oldIndex / (double)rawValues.length) * 100.0;
    int newIndex          = (int) Math.round((oldPctOfMaxIdx * newValues.length) / 100.0); 
    
    if (newIndex > 0) newIndex--;
    
    attributes[atrNo].setSimpleValue(newValues[newIndex]);
  }
//---------------------------------------------------------------------//
  public void simplifyRealValues(Dataset ds, int nValues_new, int atrNo)
  {
  	double[] rawValues = ds.getRealValuesForAttributeIn(null, atrNo);
  	double origValue   = attributes[atrNo].getValue();
  	
    // we don't increase the range!
  	if (nValues_new > rawValues.length) return;
  	
  	int oldIndex = -1;
  	for (int i=0; i < rawValues.length; i++)
  	{
  		if (origValue == rawValues[i])
  		{
  			oldIndex = i+1; // to avoid having to write "rawValues.length-1" 
  			break;
  		}
  	}
  	
    double[] newValues    = ds.reduceDiversityReal(rawValues, nValues_new, atrNo);   
    double oldPctOfMaxIdx = ((double)oldIndex / (double)rawValues.length) * 100.0;
    int newIndex          = (int) Math.round((oldPctOfMaxIdx * newValues.length) / 100.0); 
    
    if (newIndex > 0) newIndex--;
    
    /*
    if (atrNo == 0 && ds.getName().startsWith("Glass"))
    {
    	// System.out.println("---------------------------------------");
    	// for (int i=0; i < newValues.length; i++) System.out.println("" + newValues[i]);
    	// for (int i=0; i < rawValues.length; i++) System.out.println("" + rawValues[i]);
    	// System.exit(0);
    }
    */
    
    attributes[atrNo].setRealValue(newValues[newIndex]);
  }
//---------------------------------------------------------------------//
  /**
   * In case we want to use the 'raw' value of an attribute
   * for its 'simpleValue' too... Can ONLY be used if the original value
   * is of type INTEGER! 
   */
  public void setSimplevalue_directcopy(int attributeNo)
  {
    attributes[attributeNo].setSimpleValue((int)attributes[attributeNo].getValue());
  }
//---------------------------------------------------------------------//
  /**
   * Sets the attribute-values for each 'word', the value being the bitmask
   * for all possible POS classes it can be, (and any other info implemented)
   */
 public void createAttributeValues_NLP(Lexicon lex)
 {
   int[] bitmask;
   
   for (int i=0; i < attributes.length; i++)
   {
     bitmask = lex.getLexemeBitmaskFor(attributes[i].getValueTxt());
     
     if (bitmask[1] != -1)
     {
       attributes[i].setSimpleValue(bitmask[0]);
       attributes[i].setRealValue(bitmask[0]);
     }
     else
     {
       System.out.println("Lexeme " + attributes[i].getValueTxt() + " NOT found in lexicon!");
     }
   }
 }
//---------------------------------------------------------------------//
 public void setPlainANN_data(double[] pattern)
 {
    annData = pattern; 
 }
//---------------------------------------------------------------------//
  /**
   * If all values are in range:   {0.0 -> 1.0} there is no need to normalize anything.
   * Otherwise, normalize by dividing all values by the highest value.
   * 
   * In case of any negative attributes, add the 'atrMin[curAttribute]' to all attributes
   * to shift all values an equal amount upwards; this is done BEFORE the divide-normalizing!
   */
  public final void create_plain_ANN_format(double atrMax, double atrMin)
  {
    annData = new double[attributes.length];
    
    for (int i=0; i < attributes.length; i++)
    {
      if (attributes[i].hasNumericValue())
      {
        annData[i] = attributes[i].getValue();
      }
      else
      {
        byte[] bytes = attributes[i].getValueTxt().toUpperCase().getBytes();
        String byteStr = "";
        
        for (int j=0; j < bytes.length; j++) byteStr += bytes[j];        
        byteStr = "0." + byteStr;
        
        annData[i] = Double.parseDouble(byteStr);
      }
    }

    double globalMin = atrMin;    
    double globalMax = atrMax;
    
    ///// if we have any negative values in our dataset;
    if (globalMin < 0.0) fixNegativeValues(globalMin);
    
    ///// if we have any values > 1.0 in our dataset;
    if (globalMax > 1.0) normalizeValues(globalMax);
  }
//---------------------------------------------------------------------//
  /**
   * Basically same as 'createExpandedPattern()' below, but;
   * 'patternWidth' will be # of words in dictionary(mode 1), and each
   * bitpattern will only reflect which WORD NUMBER (id) 
   * the attribute consists of. No POS information!
   * 
   * Param 'nBits' are only used when creating compact-patterns!
   * Mode 2: using compactpattern the word # is directly converted to
   * its binary representation; word #7 will be 111, etc.
   */
  public void createSimpleExpandedPattern(Lexicon lex, int nBits, boolean compactPattern)
  {
    expandedPattern_NO_INFO = new ExpandedPattern[attributes.length];

    int wordNo;
    double[] pat;
    String binStr;
    
    if (compactPattern)
    {
      for (int i=0; i < attributes.length; i++)
      {
        pat    = new double[nBits];
        wordNo = lex.getLexemeIDFor(attributes[i].getValueTxt());     
        binStr = Integer.toBinaryString(wordNo);

        //  make SURE pattern is zeroed out
        for (int j=0; j < pat.length; j++) pat[j] = 0.0;
        
        // fill in from right to left
        for (int j=(binStr.length()-1); j >= 0; j--)
        {
          if (binStr.charAt(j) == '0') pat[j] = 0.0;
          else                         pat[j] = 1.0;
        }
        
        expandedPattern_NO_INFO[i] = new ExpandedPattern(pat);
      }
    }
    else
    {
      int patternWidth = lex.getWordsInDictionary().size();
      
      for (int i=0; i < attributes.length; i++)
      {
        wordNo = lex.getLexemeIDFor(attributes[i].getValueTxt());     
        pat    = new double[patternWidth];

        // make SURE pattern is zeroed out
        for (int j=0; j < patternWidth; j++) pat[j] = 0.0;
        
        pat[wordNo]                = 1.0;
        expandedPattern_NO_INFO[i] = new ExpandedPattern(pat);
      }
    }
  }
//---------------------------------------------------------------------//
  /**
   * Assumes the value (integer) "Attribute.getSimpleValue()" is already set
   * and this value should be encoded as a bitmask; that is, the integer
   * is generated by OR'ing together of POS-bits.
   * This method uses this bitmask to expand it to 'patternWidth' bits, creating
   * a double['patternwith'] array for each attribute.
   * 
   * (Each instance has different # of attribs, because instance == sentence, and
   * attribs == words). 
   */
  public void createExpandedPattern(int patternWidth)
  {
    long bitMask;
    String padZ, bitStr;
    expandedPattern_POS_INFO = new ExpandedPattern[attributes.length];
    
    for (int i=0; i < attributes.length; i++)
    {
      bitMask = attributes[i].getSimpleValue();      
      bitStr  = Long.toBinaryString(bitMask);
      padZ    = "";
      
      for (int j=0; j < (patternWidth - bitStr.length()); j++)
      {
        padZ += "0";
      }
      
      bitStr  = padZ + bitStr;  
      expandedPattern_POS_INFO[i] = new ExpandedPattern(bitStr);
    }
  }
//---------------------------------------------------------------------//
  /**
   * For the ANN-classifier:  
   *    - If any values  are negative, add the abs() of the most negative 
   *      value to EVERY attribute value in the dataset! 
   */
  private final void fixNegativeValues(double globalAtrMin)
  {
    for (int i=0; i < annData.length; i++)
    {
      annData[i] += Math.abs(globalAtrMin);
    }
  }
//---------------------------------------------------------------------//
  /**
   * For the ANN-classifier:
   *   - Normalize all values to the range {0.0 - 1.0}
   * 
   *  Divide every value by the global maximum value.
   */
  private final void normalizeValues(double globalAtrMax)
  {
    for (int i=0; i < annData.length; i++)
    {
      annData[i] /= globalAtrMax;
    }
  }
//---------------------------------------------------------------------//  
  public final Attribute[] getAttributes()
  {
    return attributes;
  }
//---------------------------------------------------------------------//  
  public final Attribute getAttribute(int atrNo)
  {
    return attributes[atrNo];
  }
//---------------------------------------------------------------------//
  public final double[] getPlainNeuralNetData()
  {
    return annData;
  }
//---------------------------------------------------------------------//
  /**
   * This returns the converted data, as above, but in an extended format:
   * Here, each attribute, (or element), of an instance is itself a bit-pattern,
   * most likely in a binary-form of 1 or 0, as opposed to the normal real-valued
   * versions of the individual attributes.... 
   * 
   * This is especially useful for language-parsing, where all attributes are textual,
   * and where we use a SERIAL input scheme: instead of just one input node, we get a
   * richer pattern-input, thus increasing pattern variance and facilitating pattern-
   * discrimination
   */
  public final ExpandedPattern[] getExpandedNeuralNetData_POS_INFO()
  {
    return expandedPattern_POS_INFO;
  }
//---------------------------------------------------------------------//
  /**
   * Pattern is simply a n-bit pattern, containing only UID info;
   * that is, the word number, ranging from 0 to # of words in dictionary
   */
  public final ExpandedPattern[] getExpandedNeuralNetData_NO_INFO()
  {
    return expandedPattern_NO_INFO;
  }
//---------------------------------------------------------------------//
  public int getClassID()
  {
    return classID;
  }
//---------------------------------------------------------------------//
  public final String getLabel()
  {
    return label;
  }
//---------------------------------------------------------------------//
}