package sharkey_mod_nlp.dataset.input;

import java.util.*;
import sharkey_mod_nlp.dataset.*;
import sharkey_mod_nlp.treemodel.*;
import sharkey_mod_nlp.gui.tree.*;

/**
 * This is a general parser for the datasets.
 * Should be called by a 'dataset loader' specific for each dataset.
 * ::-> Needs to know how the attributes are separated, and the names the attributes.
 */
public abstract class Parser
{
  private static int tmpClassID = -1;
  private static int instanceName_AttribNo = -1;
//--------------------------------------------------------------//
  /**
   * Define which attribute that will be used to 'name' each instance in
   * the dataset. If there are no such attribute, plain numbering will be used instead...
   * (::-> instanceName_AttribNo == -1)
   */
  public static void setInstanceName_AttributeNumber(int attribNo)
  {
    instanceName_AttribNo = attribNo;
  }
//--------------------------------------------------------------//
  /**
   * Takes a general dataset, stored as lines,
   * with attribute boundaries defined by 'separator'
   * and parses the data into the general 'Dataset' container.
   * 
   * Assumes complete datasets; missing data(attributes) are not handled. 
   */
  public static void parseData(Dataset dataset, Vector data, String separator)
  {
    String line;
    String[] attributes;
    String instanceLabel;
    
    for (int i=0; i < data.size(); i++)
    {
      line            = (String) data.elementAt(i);
      attributes      = getAttributes(line, separator);  
      
      if (instanceName_AttribNo != -1)
      {
        instanceLabel = attributes[instanceName_AttribNo]; 
      }
      else
      {
        instanceLabel = "" + i;
      }
      
      dataset.addInstance(attributes, tmpClassID, instanceLabel);
      tmpClassID = -1;
    }
  }
  
//--------------------------------------------------------------//
  public static void parsePOS(Dataset dataset, Vector data, String separator)
  {
    int id = 0;
    String line;
    String[] attributes;
    
    for (int i=0; i < data.size(); i++)
    {
      line            = (String) data.elementAt(i);
      attributes      = getAttributes(line, separator);  
      dataset.addPOS(attributes, id++);
    }
  }
//--------------------------------------------------------------//
  public static void parseDIC(Dataset dataset, Vector data, String separator, int patternWidth)
  {
    int id = 0;
    String line;
    String[] attributes;
    
    for (int i=0; i < data.size(); i++)
    {
      line            = (String) data.elementAt(i);
      attributes      = getAttributes(line, separator);  
      dataset.addLexiconEntry(attributes, id++, patternWidth);
    }
  }
//--------------------------------------------------------------//
  public static void parsePreStructuredInput(Dataset dataset, Vector data, String separator)
  {
    int id = 0;
    String line;
    String[] attributes;
    Lexicon lex = dataset.getLexicon();
    
    for (int i=0; i < data.size(); i++)
    {
      line            = (String) data.elementAt(i);
      attributes      = getAttributes(line, separator);  
      id              = Integer.parseInt(attributes[0]); 
      ParseNode root  = parseStructure(attributes[1], lex, i);
      
      root.setSyntacticStructureID(id); // AKA classID, but specific to the RAAM
      dataset.addParseStructure(root);
    }
  }
//--------------------------------------------------------------//
  /**
   * 1. Build a tree, using pre-structured input. (braces defines units)
   * 2. Find which elements (POS) goes together to create parentNodes;
   *    this determines which elements to give as input to the RAAM.
   * 3. Autoassociate them -> store the hiddenRep. values in their parentnode... 
   * The initial input-patterns are the 8 different syntactic structures 
   * found in the dataset.
   */
  private static ParseNode parseStructure(String s, Lexicon lex, int structNo)
  {
    Stack<ParseNode> stack = new Stack<ParseNode>();    
    ParseNode parent       = new ParseNode(null, "S");
    ParseNode root         = parent;

    int currentDepth = 0, elemCount = 0;
    
    parent.setTextualValue(new String[] {"S"});
    parent.setPly(0);
    
    // There may NEVER be more than 1 '(' in direct succession!
    // (no restrictions on number of closing ')'.
    
    for (int i=0; i < s.length(); i++)
    {
      if (s.charAt(i) == ' ')
      {
        continue;
      }
      else if (s.charAt(i) == '(')
      { 
        // Create an empty node, as container for one of the phrase-nodes { NP, VP, AP, PP, etc }
        ParseNode pn = new ParseNode(parent, "xXx-"+i);
        pn.setTextualValue(new String[] {"xP"});
        pn.setPly(parent.getPly() + 1);
        parent.addChild(pn);
        
        stack.push(parent);
        parent = pn;
        
        currentDepth++;  
        continue; 
      }      
      else if (s.charAt(i) == ')')
      {
        if (!stack.isEmpty()) parent = stack.pop();
        
        currentDepth--;
        continue;
      }
      
      // each 'terminal' element are added to current parent
      // ... get next element
      String element = nextElement(s, i);
      if (element != null)
      {
        int posID = lex.getPosIDFor(element.trim().toUpperCase());
        
        ParseNode pn = new ParseNode(parent, element);
        pn.setPOS_ID(posID);
        pn.setBitpattern(Dataset.getExpandedPatternFromInt(posID, 12));        
        pn.setBitNo(ParseNode.BITNO_TERMINAL, 1.0);
        pn.setTextualValue(new String[] {element, elemCount+""});
        pn.setPly(parent.getPly() + 1);
        parent.addChild(pn);
        // pn.invertBitPattern();
        
        /*
        System.out.print("pattern for termNode : ");
        double[] bp = pn.getBitpattern();
        for (int x=0; x < pn.getBitpattern().length; x++) System.out.print(bp[x] + " ");
        System.out.print(" POS = " + pn.getNodeLabel() + ", " + posID); System.out.println();
        */
        
        i += (element.length() -1); // -1 because of 'i++' in for-loop body...
        elemCount++;
      }      
    }
    
    
    ////////////////////////////////////////////////////
    ///// debug by showing parse-result as tree
    if (structNo == -1) // show tree
    {
      Spatializer spz   = new Spatializer(new Tree<IGenericNode>(root), 45, 45, 14, 14);
      TreeCanvas canvas = new TreeCanvas(spz, false);
      TreeFrame tf      = new TreeFrame("Parsetree", canvas);
      tf.setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
      tf.setVisible(true);
    }
    
    // there's probably someting wrong with the parsing - or the structure definition from file...
    if (currentDepth != 0) System.out.println("unbalanced Ply! : " + currentDepth);
    
    return root;
  }
//--------------------------------------------------------------//
  private static String nextElement(String s, int idx)
  {
    char c;
    
    for (int i=idx; i < s.length(); i++)
    {
      c = s.charAt(i);
      
      if (c == ' ' || c == '(' || c == ')')
      {
        return s.substring(idx, i);
      }
    }    
    return null;
  }
//--------------------------------------------------------------//
  private static String[] getAttributes(String instanceLine, String separator)
  {
    StringTokenizer tok = new StringTokenizer(instanceLine, separator);
    Vector<String> v    = new Vector<String>();    
    String token        = null;
    
    while (tok.hasMoreTokens())
    {
      token = tok.nextToken().trim();
      token = removeIllegals(token);
      
      // elements prefixed with '@' are to be ignored; they are NOT part of the feature-data.
      if (!token.startsWith("@")) v.addElement(token);
    }

    // CLASS ID IS ASSUMED TO BE THE LAST ONE!
    try
    {
      token = token.replace("@", "");
      tmpClassID = Integer.parseInt(token);
    }
    catch (NumberFormatException e) { }
    
    ///////////////////////////////////////////////////////////////////
    // create the attribute-array containing only the valid attribs...
    String[] attribs = new String[v.size()];
    for (int i=0; i < v.size(); i++) 
    {
      attribs[i] = v.elementAt(i);
    }
    
    return attribs;
  }
//--------------------------------------------------------------//
  private static String removeIllegals(String str)
  {
    byte[] legalBytes_lower = "abcdefghijklmnopqrstuvwxyz 0123456789,.-+*!%#@$&()={}[]<>_/\\".getBytes();
    byte[] legalBytes_upper = "abcdefghijklmnopqrstuvwxyz".toUpperCase().getBytes();
    byte[] bytes      = str.getBytes();
    String pureStr    = "";
    boolean ok;
    
    for (int i=0; i < bytes.length; i++)
    {
      ok = false;
      for (int j=0; j < legalBytes_lower.length; j++)
      {
        if (bytes[i] == legalBytes_lower[j])
        {
          ok = true;
          break;
        }
      }
      
      for (int j=0; j < legalBytes_upper.length; j++)
      {
        if (bytes[i] == legalBytes_upper[j])
        {
          ok = true;
          break;
        }
      }
      
      if (ok) pureStr += (char)bytes[i];
    }
    
    int cmntIdx = pureStr.indexOf("#");
    if (cmntIdx != -1)
    {
      pureStr = pureStr.substring(0,cmntIdx);
    }
    
    return pureStr;
  }
//--------------------------------------------------------------//
}
