package wang.codec;

import java.util.*;

public final class Network
{
  public static final int DECODER = 0x00; 
  public static final int ENCODER = 0x01;
  
  private Node[] inputLayer, outputLayer;
  private double[] outputStorage;
//--------------------------------------------------------------//  
  public Network(int inputDim, int outputDim)
  {
    inputLayer  = new Node[inputDim];
    outputLayer = new Node[outputDim];
    
    outputStorage = new double[outputDim];

    //////////////////////
    //// Create the links
    Link[] links = new Link[inputDim * outputDim];
    for (int i=0; i < links.length; i++)
    {
      double rnd = (Math.random() * 2) - 1.0;
      links[i]   = new Link(rnd, false); 
    }

    ///////////////////////
    //// Create the nodes
    for (int i=0; i < inputDim; i++)  inputLayer[i]  = new Node();
    for (int i=0; i < outputDim; i++) outputLayer[i] = new Node();
    
    ////////////////////////
    //// Setup the topology
    int cnt = 0;
    for (int i=0; i < inputDim; i++)
    {
      for (int j=0; j < outputDim; j++)
      {
        links[cnt].setInNode(inputLayer[i]);
        links[cnt].setOutNode(outputLayer[j]);
        
        inputLayer[i].addLink(links[cnt]);
        outputLayer[j].addLink(links[cnt]);
        
        cnt++;
      }
    }
  }
//--------------------------------------------------------------//
  public final void updateWeight(int node_in, int node_out, double dWt)
  {
    Link link = inputLayer[node_in].getLink(node_out);
    link.updateWeight(dWt);
  }
//--------------------------------------------------------------//
  public final void setInputPattern(double[] inputPattern)
  {
    for (int i=0; i < inputLayer.length; i++)
    {
      inputLayer[i].addInput(inputPattern[i]);
    }
  }
//--------------------------------------------------------------//  
  public final void propagate()
  {
    for (int i=0; i < inputLayer.length; i++) inputLayer[i].propagateInputForward();    
    for (int i=0; i < outputLayer.length; i++) outputStorage[i] = outputLayer[i].getInput();
  }
//--------------------------------------------------------------//  
  public final double[] getOutputs()
  {
    return outputStorage;
  }
//--------------------------------------------------------------//
  /**
   * Resets all inputs in all nodes to 0.0. The weights are NOT changed.
   */
  public final void zeroAllInputData()
  {
    for (int i=0; i < inputLayer.length; i++)  inputLayer[i].reset(0.0);
    for (int i=0; i < outputLayer.length; i++) outputLayer[i].reset(0.0);
  }
//--------------------------------------------------------------//  
}



//---------------------------------------------//
//******************************************* //
//---------------------------------------------//

final class Link
{
  private static boolean enforceWeightBoundary;

  private double weight, lowBound, hiBound;
  private Node inNode, outNode;
  
  public Link(double wt, boolean enforceWtBounds)
  {
    this.weight           = wt;
    enforceWeightBoundary = enforceWtBounds;
    inNode                = null;
    outNode               = null;
  }
//--------------------------------------------------------------//  
  public final void setBounds(double hi, double low)
  {
    this.hiBound  = hi;
    this.lowBound = low;
  }
//--------------------------------------------------------------//
  public final void setInNode(Node n)
  {
    this.inNode = n;
  }
//--------------------------------------------------------------//
  public final void setOutNode(Node n)
  {
    this.outNode = n;
  }
//--------------------------------------------------------------//
  public final Node getInNode()
  {
    return this.inNode;
  }
//--------------------------------------------------------------//
  public final Node getOutNode()
  {
    return this.outNode;
  }
//--------------------------------------------------------------//
  public final double weight()
  {
    return this.weight;
  }
//--------------------------------------------------------------//
  public final void updateWeight(double dWt)
  {
    if (enforceWeightBoundary)
    {
      if (((weight + dWt) >= lowBound) && ((weight + dWt) <= hiBound))
      {
        this.weight += dWt;
      }
    }
    else
    {
      this.weight += dWt;
    }
  }
}

//---------------------------------------------//
// ******************************************* //
//---------------------------------------------//

final class Node
{
  private double net_input;
  private Vector links;
  
  public Node()
  {
    this.net_input = 0.0;
    links          = new Vector();
  }
//--------------------------------------------------------------//
  public final void propagateInputForward()
  {
    Link tmpLink;
    
    for (int i=0; i < links.size(); i++)
    {
      tmpLink = (Link) links.elementAt(i);      
      tmpLink.getOutNode().addInput(net_input * tmpLink.weight());
    }
  }
//--------------------------------------------------------------//
  @SuppressWarnings("unchecked")
  public final void addLink(Link lnk)
  {
    links.addElement(lnk);
  }
//--------------------------------------------------------------//
  public final void addInput(double input)
  {
    net_input += input;
  }
//--------------------------------------------------------------//
  public final Vector getLinks()
  {
    return links;
  }
//--------------------------------------------------------------//
  public final Link getLink(int linkNo)
  {
    return (Link)links.elementAt(linkNo);
  }
//--------------------------------------------------------------//
  public final double getInput()
  {
    return net_input;
  }
//--------------------------------------------------------------//
  public final void reset(double defValue)
  {
    net_input = defValue; // usually ZERO
  }
//--------------------------------------------------------------//
}