package wang.env;

import wang.codec.Network;

import java.util.*;

public final class Agent implements Runnable
{
  public static final int ROLE_SPEAKER  = 0x00;
  public static final int ROLE_LISTENER = 0x01;
  private static int counter = 0x00;
  private static boolean strictlyTurnbased;
  private static int curEpoch = 0x00;
  
  private Environment env;
  private Thread myThread;
  private Vector messageBuffer;
  private int activeRole, agentID, meaningPtr;
  private boolean stop;
  private String name;
  
  private Network encoder, decoder;
//--------------------------------------------------------//    
  public Agent(String name, int role, Environment env)
  {
    activeRole     = role;
    agentID        = Agent.counter++;
    messageBuffer  = new Vector();
    this.name      = name;
    this.env       = env;
    meaningPtr     = 0;
    
    strictlyTurnbased = false; // default to free 'chaotic' agent behaviour
  }
//--------------------------------------------------------//
  public void initCodecs(int n_meanings, int n_cWords)
  {
    encoder = new Network(n_meanings, n_cWords);
    decoder = new Network(n_cWords, n_meanings);
  }
//--------------------------------------------------------//
  /**
   * Note: 'communication' here would also include any action requiring 
   * feedback from another agent.
   * Enforce turnbased communication - synchronizing actions
   * such that an agent does not initiate any new actions before its
   * current (message) has been processed by target agent(s). This guarantees
   * that the connection strengths are updated after each communication attempt.
   * 
   *  - free, unsynchronized communication would make inital training/evaluation
   *  of the W/M association codecs very messy and unreliable... in our simple case.
   */
  public final static void setTurnbasedCommunication(boolean commFlag)
  {
    Agent.strictlyTurnbased = commFlag;
  }
//--------------------------------------------------------//
  public final static int getCurrentEpoch()
  {
    return Agent.curEpoch;
  }
//--------------------------------------------------------//
  public final int getID()
  {
    return agentID;
  }
//--------------------------------------------------------//  
  public final void startAgent()
  {
    messageBuffer.removeAllElements();
    meaningPtr     = 0;
    Agent.curEpoch = 0;
    
    stop = false;
    myThread = new Thread(this);
    myThread.start();
  }
//--------------------------------------------------------//  
  public final void stopAgent()
  {
    stop = true;
  }
//--------------------------------------------------------//  
  public final void changeRole(int newRole)
  {
    this.activeRole = newRole;
  }
//--------------------------------------------------------//
  /**
   * this method is called by the environment to insert a message/stimuli
   * into an agent's internal (message) buffer.
   */
  @SuppressWarnings("unchecked")
  public final boolean giveMessage(Agent origin, int codeWord, int intendedMeaning_id)
  {
    if (Agent.strictlyTurnbased && !messageBuffer.isEmpty())
    {
      // System.out.println(name + " was BUSY!");
      return false;
    }
    
    messageBuffer.addElement(new Msg(origin, codeWord, intendedMeaning_id));
    return true;
  }
//--------------------------------------------------------//
  /**
   * This is the DEcoding of a message: mapping a codeWORD to a MEANING;
   * each of the OUTPUT-nodes in the DECODER represent a meaning.
   */
  private final boolean decodeMessage()
  {
    if (messageBuffer.isEmpty()) return false;
    
    Msg msg = (Msg) messageBuffer.remove(0);
    
    // create an input pattern for the codeword: a fixed-size input vector
    // where only 1 element are enabled / set to 1...
    double[] pattern = new double[env.getNumberOfWords()];
    createPattern(pattern, msg.codewordID);
          
    decoder.setInputPattern(pattern);
    decoder.propagate();
    double[] output = decoder.getOutputs();
    decoder.zeroAllInputData();
    
    //////////////////////////////////////////////////////////////////////////////////
    //// find which outputNode is highest; this is the interpreted meaning of the 
    //// given codeword. If this is the SAME as actually meant by the sender, update
    //// connection-weights supporting this association. Otherwise, decrease all weights
    //// that lead to this 'miscommunication'.
    int decodedMeaningIdx = findWinnerNode(output);
    double weightUpdate   = -1.0; // default to punishment :)
    
    if (decodedMeaningIdx == msg.intentedMeaningID)
    {
      // strengthen the connections in decoder that gives this word->meaning mapping
      weightUpdate = 1.0;
      env.updateStats(curEpoch);
    }      
    
    decoder.updateWeight(msg.codewordID, decodedMeaningIdx, weightUpdate);
    msg.originAgent.getEncoder().updateWeight(msg.intentedMeaningID, msg.codewordID, weightUpdate);
    
    return true;
  }
//--------------------------------------------------------//
  private final boolean encodeMessage()
  {
    if (meaningPtr >= env.getNumberOfMeanings())
    {
      Agent.curEpoch++;
      meaningPtr = 0;      
      // System.out.println("-------end of epoch " + curEpoch + " ---------");
    }
    
    double[] pattern = new double[env.getNumberOfMeanings()]; 
    createPattern(pattern, meaningPtr);
    
    encoder.setInputPattern(pattern);
    encoder.propagate();
    double[] output = encoder.getOutputs();
    encoder.zeroAllInputData();
    
    int codeWord = findWinnerNode(output);
    
    /////////////////////////////////////////////////
    //// For concistency, don't update 'meaningPtr'
    //// if the communication fails; usually because the other agent has
    //// not yet finished processing the previous message.
    if (env.talkTo(this, -1, codeWord, meaningPtr))
    {
      meaningPtr++;
      return true;
    }
    
    // recipient was busy; we did not send any message
    return false;
  }
//--------------------------------------------------------//
  /**
   * the param 'id' will be used as index of which inputNode to enable:
   * In the current codec scheme by Wang, ONLY ONE input are to be active
   * for any given (zero-padded) input.
   */
  private final void createPattern(double[] pattern, int id)
  {
    for (int i=0; i < pattern.length; i++)
    {
      if (i == id) pattern[i] = 1;
      else pattern[i] = 0;
    }
  }
//--------------------------------------------------------//
  /**
   * return which output-node has the highest activation; a standard
   * kWTA scheme, with k=1.
   */
  private final int findWinnerNode(double[] output)
  {
    double curMax = output[0];
    int maxIndex  = 0;
    
    for (int i=1; i < output.length; i++)
    {
      if (output[i] > curMax)
      {
        curMax   = output[i];
        maxIndex = i;
      }
    }
    return maxIndex;
  }
//--------------------------------------------------------//
  /**
   * Not quite encouraged in an Agent-framwork, but:
   * It really simplifies AND speeds up the communication-evaluation process
   * by allowing direct access to the encoder-network that created the message.
   * So; this method is called by the listener-agent that has just completed 
   * decoding a message, and needs to update its (decoder)weights accordingly.
   * Thus, by giving access to the speaker-agent's encoder, those weights can
   * be updated directly and timely, too...
   */
  public Network getEncoder()
  {
    return this.encoder;
  }
//--------------------------------------------------------//
  public Network getDecoder()
  {
    return this.decoder;
  }
//--------------------------------------------------------//
  public void run()
  {
    int endEpoch = env.getNumberOfEpochs(); 

    try 
    {  
      while (!stop && Agent.curEpoch < endEpoch)
      {
        if (activeRole == Agent.ROLE_LISTENER)
        {
          // LISTENER decodes a CODEWORD into a MEANING          
          decodeMessage();
        }
        else if (activeRole == Agent.ROLE_SPEAKER)
        {
          // SPEAKER encodes a MEANING into a CODEWORD          
          encodeMessage();
        }      
        Thread.sleep(0);
      }
    }
    catch (InterruptedException e) 
    {
      // System.out.println("What!?");
    }   
    
    System.out.println("Agent " + name + " stopped.");
  }
//--------------------------------------------------------//
}

/**
 * 
 *
 */
final class Msg
{
  Agent originAgent;
  int codewordID;
  int intentedMeaningID;
  
  public Msg(Agent originAgent, int cWordID, int meaningID)
  {
    this.codewordID        = cWordID;
    this.intentedMeaningID = meaningID;
    this.originAgent       = originAgent;
  }
}