package sfm.dataset;

import java.io.Serializable;
import java.util.*;

/**
 * Dataset hierarchy:
 * 
 *    Dataset--
 * 	          |
 *            -- Instance--
 *                        |
 *                        -- Attribute
 */
public final class Dataset implements Serializable
{
  private String[] classNames;
  private String datasetName;
  private Vector<Instance> dataset;
  private int numberOfClasses;
  
  private Vector<Instance> dsTrain, dsVal, dsTest;
  
  public Dataset(String name, int nClasses)
  {
    datasetName        = name;
    numberOfClasses    = nClasses;
    dataset            = new Vector<Instance>();
  }
//---------------------------------------------------------------------//
  public String getName()
  {
    return this.datasetName;
  }
//---------------------------------------------------------------------//
  /**
   * Sets (and replaces) the existing set of instances
   * already stored here, if any.
   * 
   * !PS!: This method may NOT be compatible with the existing
   * data-analyzing and preparation code! The instances supplied
   * here are supposed to be self-contained with regard to any (meta-)info
   * attatched to instances when loaded the normal way. 
   * (=> as when loading a dataset from file). 
   */
  public void setInstances(Vector<Instance> newInstances)
  {
    dataset = newInstances;
  }
//---------------------------------------------------------------------//
  public void addInstance(Instance instance)
  {
    dataset.addElement(instance);
  }
//---------------------------------------------------------------------//
  /**
   * returns the data as a set of instances; a list of class Instance objects
   */
  public final Vector<Instance> getInstanceData()
  {
    return dataset;
  }
//---------------------------------------------------------------------//
  /**
   * Testing-method ensuring same sets of instances
   * when developing the adabooster... (for each instantiation of it)
   */
  public void splitDatasetAndStorePartitions(double pctTrain, double pctVal)
  {
  	dsTrain = new Vector<Instance>();
  	dsVal   = new Vector<Instance>();
  	dsTest  = new Vector<Instance>();
  	
  	splitDataset(pctTrain, pctVal, dsTrain, dsVal, dsTest);
  }
//--------------------------------------------------------------------------------//
  public void getPartitionedDatasets(Vector<Instance> data_train, Vector<Instance> data_validation, Vector<Instance> data_test)
  {
  	for (int i=0; i < dsTrain.size(); i++) data_train.addElement(dsTrain.elementAt(i));
  	for (int i=0; i < dsVal.size(); i++)   data_validation.addElement(dsVal.elementAt(i));
  	for (int i=0; i < dsTest.size(); i++)  data_test.addElement(dsTest.elementAt(i));
  }
//--------------------------------------------------------------------------------//
  /**
   * While the selection of training-samples is randomized to ensure fair distribution(?), we
   * also SHOULD ensure that this will cover some instances from all CLASSES.
   *   - (I.e. the general assumption that a given dataset for training must be 'reasonably'
   * representative of the corresponding (un)limited set of real data, out there...)
   * 
   * NB: The rough method here will cause nearly all samples of some classes to be included 
   * in the training set..., if a class has very few samples relative to the total amount of
   * training samples.
   */
  private void splitDataset(double pctForTraining, double pctForValidation, Vector<Instance> data_train, Vector<Instance> data_validation, Vector<Instance> data_test)
  {
    data_train.removeAllElements();
    data_test.removeAllElements();
    if (data_validation != null) data_validation.removeAllElements();
    
    int nTrainElems      = (int) (pctForTraining   * getNumberOfInstances());
    int nValidationElems = (int) (pctForValidation * nTrainElems);
    
    // if dataset too small, use ALL samples for training... (probably just a dummy-dataset)
    if (pctForTraining >= 1.0 || getNumberOfInstances() < 25 || nValidationElems <= 1)
    {
      pctForTraining   = 1.0;
      pctForValidation = 0.0;
      nTrainElems      = getNumberOfInstances();
    }
    
    
    int elem;
    for (elem=0; elem < nTrainElems; elem++) data_train.addElement(dataset.elementAt(elem));
    
    ////////////////////////////////////////////////////////////
    ///// Use the given ratio to get samples to use for the  
    ///// validation-set: these are taken from the TRAINING DATA!
    if (nValidationElems > 0 && data_validation != null)
    {
      for (int i=0; i < nValidationElems; i++)
      {
        data_validation.addElement(data_train.remove(0));
      }
    }
    
    ////////////////////////////////////////////////////////
    ///// Now put the remaining samples into the 'test' bag
    for (int i=elem; i < dataset.size(); i++) data_test.addElement(dataset.elementAt(elem));
  }
//---------------------------------------------------------------------//
  public final void setClassNames(String[] cNames)
  {
    this.classNames = cNames;
  }
//---------------------------------------------------------------------//
  public String getClassName(int classID)
  {
    return classNames[classID];
  }
//---------------------------------------------------------------------//
  /**
   * returns the size of the dataset; the number of instances
   */
  public int getNumberOfInstances() 
  {
    return dataset.size(); 
  }
//---------------------------------------------------------------------//
  public int getNumberOfClasses() 
  { 
    return numberOfClasses; 
  }
//---------------------------------------------------------------------//
}