package sfm.dataset;

import java.io.Serializable;

//-===============================================================--//
/**
 * This is the container class for a data-sample (instance) in the dataset;
 * All attributes are stored here.
 * 
 * ~ An instance is the equvivalent of the ANN's Pattern-class; yielding 1 inputNode
 *   pr. attribute, or x nodes if we use expanded patterns: then each attribute has
 *   a n-width bitpattern. Useful for encoding more info pr. attrib when they
 *   are natural language "words". (lexemes)
 */
public final class Instance implements Serializable
{
  private static int INSTANCE_COUNTER = 0x00;
	
  private String label;
  private int classID, id;
  
  private ExpandedPattern[] expandedPattern;
  private double[] annData, trainingTarget;
  private double weight, minWeight, maxWeight;
  
  /**
   * Generation of NeuralNet values:
   * Values of all attributes are normalized into the range [0.0, 1.0].
   */  
  public Instance(String label, double[] pattern, int classID, double[] tt)
  {
    this.annData = pattern;
    this.classID = classID;
    this.label   = label;
    this.id      = INSTANCE_COUNTER++;
    
    this.trainingTarget = tt;
    
    weight     = 1.0;
    minWeight  = 1.0;
    maxWeight  = 1.0;
  }
//---------------------------------------------------------------------//
  public int getInstanceNumber()
  {
  	return id;
  }
//---------------------------------------------------------------------//
  public static void resetInstanceCounter()
  {
  	Instance.INSTANCE_COUNTER = 0x00;
  }
//---------------------------------------------------------------------//
  public double getUpperWeightBound()
  {
  	return maxWeight;
  }
//---------------------------------------------------------------------//
  public double getLowerWeightBound()
  {
  	return minWeight;
  }
//---------------------------------------------------------------------//
  public void setWeightBounds(double min, double max)
  {
  	minWeight = min;
  	maxWeight = max;
  }
//---------------------------------------------------------------------//
  public boolean changeWeight(double dWt)
  {
  	weight += dWt;
  	
    if (weight < minWeight)
    {
    	weight = minWeight;
    	return false;
    }
  	else if (weight > maxWeight)
  	{
  		weight = maxWeight;
  		return false;
  	}
  	return true;
  }
//---------------------------------------------------------------------//
  public void setWeight(double wt)
  {
  	weight = wt;
  	
  	     if (weight < minWeight) weight = minWeight;  
  	else if (weight > maxWeight) weight = maxWeight;
  }
//---------------------------------------------------------------------//
  public double getWeight()
  {
  	return weight;
  }
//---------------------------------------------------------------------//
 public void setPlainANN_data(double[] pattern)
 {
    annData = pattern; 
 }
//---------------------------------------------------------------------//
  public final double[] getPlainNeuralNetData()
  {
    return annData;
  }
//---------------------------------------------------------------------//
  public final double[] getTrainingTarget()
  {
    return this.trainingTarget;
  }
//---------------------------------------------------------------------//
  /**
   * This returns the converted data, as above, but in an extended format:
   * Here, each attribute, (or element), of an instance is itself a bit-pattern,
   * most likely in a binary-form of 1 or 0, as opposed to the normal real-valued
   * versions of the individual attributes.... 
   * 
   * This is especially useful for language-parsing, where all attributes are textual,
   * and where we use a SERIAL input scheme: instead of just one input node, we get a
   * richer pattern-input, thus increasing pattern variance and facilitating pattern-
   * discrimination
   */
  public final ExpandedPattern[] getExpandedNeuralNetData()
  {
    return expandedPattern;
  }
//---------------------------------------------------------------------//
  public int getClassID()
  {
    return classID;
  }
//---------------------------------------------------------------------//
  public final String getLabel()
  {
    return label;
  }
//---------------------------------------------------------------------//
}