package sfm.shp;

import java.util.*;
import java.awt.image.*;
import java.awt.*;
import java.awt.geom.*;
import sfm.dataset.*;

/**
 *  All SHAPES are considered pure 2D, and single color - for easy differentiation
 *  between shape pixels and background. This ShapeData structure is to be used only
 *  for learning SHAPES, ignoring properties such as color, size, etc.
 * 
 * Containing data for -
 * 
 *   1. Edge pixels; pixels making up the outer bounds of the shape, the circumference
 *   2. All pixels in shape, both edge pix and all of the 2D surface pixels...
 *   3. Width and height of shape, in terms of a bounding box
 *   4. the imagesource, as a BufferedImage object.
 *   
 *   each pixel-data consists of 3 elements: x, y, colorvalue
 *
 */
public class ShapeData
{
  private static Vector<ShapeData> shapePool = new Vector<ShapeData>();
  
  private static int SHP_CNT = 0x00;
  private static int globalMaxElementsInShape = 0;
  
  // Default colors for background and shape: black & white
  private static int BACKGR_R     = 0;
  private static int BACKGR_G     = 0;
  private static int BACKGR_B     = 0;
  private static int SHP_R        = 255;
  private static int SHP_G        = 255;
  private static int SHP_B        = 255;
  private static int MAX_VARIANCE = 25;
  
  private BufferedImage imgObj;
  private String name;
  private int shpID; // can be used as classID...
  private Vector<PixelData> pixelV, edgePixelV;
  // private Polygon polygon_filled, polygon_edge;
  
  private double maxX, maxY;
  private double[] pattern, tt;
  
  public ShapeData(String shpName, BufferedImage imgSrc)
  {
    name       = shpName.toUpperCase();
    pixelV     = new Vector<PixelData>();
    edgePixelV = new Vector<PixelData>();
    imgObj     = imgSrc;
    shpID      = SHP_CNT++;
    pattern    = null;
    
    extractShapeData();
    extractEdgePixels();
    minimizeCoordinates();
    
    // createPolygon(pixelV);
    // createPolygon(edgePixelV);
    
    if (edgePixelV.size() > globalMaxElementsInShape)
    {
      globalMaxElementsInShape = edgePixelV.size(); 
    }
    
    shapePool.addElement(this);
  }
//------------------------------------------------------------------//
  public static final Vector<ShapeData> getShapePool()
  {
    return shapePool;
  }
//------------------------------------------------------------------//
  public static int getPatternLength(int componentsInEachElement)
  {
    return globalMaxElementsInShape * componentsInEachElement;
  }
//------------------------------------------------------------------//
  public static void specifyColorSlack(int slack)
  {
    ShapeData.MAX_VARIANCE = slack;
  }
//------------------------------------------------------------------//
  public static void specifyBackgroundColor(int r, int g, int b)
  {
    ShapeData.BACKGR_R = r;
    ShapeData.BACKGR_G = g;
    ShapeData.BACKGR_B = b;
  }
//------------------------------------------------------------------//
  public static void specifyShapeColor(int r, int g, int b)
  {
    ShapeData.SHP_R = r;
    ShapeData.SHP_G = g;
    ShapeData.SHP_B = b;
  }
//------------------------------------------------------------------//
  public int getNumberOfPixelsInShapeSurface()
  {
    return pixelV.size();
  }
//------------------------------------------------------------------//
  public int getNumberOfPixelsInShape()
  {
    return edgePixelV.size();
  }
//------------------------------------------------------------------//
  private void XcreatePolygon(Vector<PixelData> pixelVx)
  {
    /*
    int[] xp = new int[pixelVx.size()];
    int[] yp = new int[pixelVx.size()];
    
    for (int i=0; i < pixelVx.size(); i++)
    {
      xp[i] = pixelVx.elementAt(i).getX();
      yp[i] = pixelVx.elementAt(i).getY();
    }
    
    if (pixelVx.size() == pixelV.size()) polygon_filled = new Polygon(xp, yp, xp.length);
         
    else if (pixelVx.size() == edgePixelV.size())
    {
      polygon_edge = new Polygon(xp, yp, xp.length);
    }
    */
  }
//------------------------------------------------------------------//
  private void extractShapeData()
  {
    Color c;
    int r, g, b;
    
    for (int i=0; i < imgObj.getHeight(); i++)
    {
      for (int j=0; j < imgObj.getWidth(); j++)
      {
        c = new Color(imgObj.getRGB(j, i));
        r = c.getRed();
        g = c.getGreen();
        b = c.getBlue();
        
        if (Math.abs(r - SHP_R) <= MAX_VARIANCE)
        {
          if (Math.abs(g - SHP_G) <= MAX_VARIANCE)
          {
            if (Math.abs(b - SHP_B) <= MAX_VARIANCE)
            {
              pixelV.addElement(new PixelData(j,i, imgObj.getRGB(j, i)));
            }
          }
        }
        //---endif
      }
    }
  }
//------------------------------------------------------------------//
  private void extractEdgePixels()
  {
    int x,y;
    
    maxX = -1.0;
    maxY = -1.0;
    
    for (int i=0; i < pixelV.size(); i++)
    {
      x = pixelV.elementAt(i).getX();
      y = pixelV.elementAt(i).getY();
      
      if (isEdgePixel(x,y))
      {
        edgePixelV.addElement(pixelV.elementAt(i));
        
        if (x > maxX) maxX = x;
        if (y > maxY) maxY = y;
      }
    }
  }
//------------------------------------------------------------------//
  /**
   * ShapeData in 'pixelV' is assumed to contain ONLY pixels on or in
   * the shape it represents. I.e. no boundingbox repr.
   * A pixel(x,y) is considered to be an EDGEPIXEL if it is NOT
   * surrounded by a pixel on each of the 8 points around it. 
   */
  private boolean isEdgePixel(int x, int y)
  {
    PixelData pixD;
    Point W,E,N,S,NW,NE,SW,SE;
    int surrCnt = 0;

    W = new Point(x-1, y);
    E = new Point(x+1, y);
    N = new Point(x, y-1);
    S = new Point(x, y+1);
    NW = new Point(x-1, y-1);
    NE = new Point(x+1, y-1);
    SW = new Point(x-1, y+1);
    SE = new Point(x+1, y+1);
    
    for (int i=0; i < pixelV.size(); i++)
    {
      pixD = pixelV.elementAt(i);
      
           if ((pixD.getX() == W.getX())  && (pixD.getY() == W.getY()))  surrCnt++;
      else if ((pixD.getX() == E.getX())  && (pixD.getY() == E.getY()))  surrCnt++;
      else if ((pixD.getX() == N.getX())  && (pixD.getY() == N.getY()))  surrCnt++;
      else if ((pixD.getX() == S.getX())  && (pixD.getY() == S.getY()))  surrCnt++;
      else if ((pixD.getX() == NW.getX()) && (pixD.getY() == NW.getY())) surrCnt++;
      else if ((pixD.getX() == NE.getX()) && (pixD.getY() == NE.getY())) surrCnt++;
      else if ((pixD.getX() == SW.getX()) && (pixD.getY() == SW.getY())) surrCnt++;
      else if ((pixD.getX() == SE.getX()) && (pixD.getY() == SE.getY())) surrCnt++;
    }
    
    return (surrCnt != 8);
  }
//------------------------------------------------------------------//
  
//------------------------------------------------------------------//
  private void minimizeCoordinates()
  {
    PixelData pixD;
    int minX = pixelV.elementAt(0).getX();
    int minY = pixelV.elementAt(0).getY();
    
    for (int i=1; i < pixelV.size(); i++)
    {
      pixD = pixelV.elementAt(i);
      
      if (pixD.getX() < minX) minX = pixD.getX();
      if (pixD.getY() < minY) minY = pixD.getY();
    }
    
    for (int i=0; i < pixelV.size(); i++)
    {
      pixD = pixelV.elementAt(i);
      pixD.changeX(-minX);
      pixD.changeY(-minY);
    }
  }
//------------------------------------------------------------------//
  public double[] getTrainingTarget()
  {
    tt = new double[SHP_CNT];
    
    for (int i=0; i < tt.length; i++)
    {
      tt[i] = 0.0;
    }
    
    tt[shpID] = 1.0;
    
    return tt;
  }
//------------------------------------------------------------------//
  public Vector<PixelData> getShapeData()
  {
    return pixelV;
  }
//------------------------------------------------------------------//
  public Vector<PixelData> getEdgePixels()
  {
    return edgePixelV;
  }
//------------------------------------------------------------------//
  public double[] getPattern(boolean equalizeLength)
  {
    if (pattern != null) return pattern;

    
    //  patttern consists of two components, x,y - so we multiply be 2    
    if (equalizeLength) pattern = new double[globalMaxElementsInShape * 2];
    else                pattern = new double[edgePixelV.size() * 2];
    
    double x,y;
    int idx = 0;
    
    for (int i=0; i < pattern.length; i++)
    {
      pattern[i] = 0.0; 
    }
    
    for (int i=0; i < edgePixelV.size(); i++)
    {
      x = edgePixelV.elementAt(i).getX() / maxX;
      y = edgePixelV.elementAt(i).getY() / maxY;
      
      pattern[idx+0] = x;
      pattern[idx+1] = y;
      
      idx += 2;
    }
    
    return pattern;
  }
//------------------------------------------------------------------//
  public Vector<double[]> getExpandedPattern()
  {
    Vector<double[]> xpV = new Vector<double[]>();
    double x,y;
    
    for (int i=0; i < edgePixelV.size(); i++)
    {
      x = edgePixelV.elementAt(i).getX() / maxX;
      y = edgePixelV.elementAt(i).getY() / maxY;
      
      xpV.addElement(new double[] {x,y});
    }
    
    return xpV;
  }
//------------------------------------------------------------------//
  public BufferedImage getShapeAsImage()
  {
    return imgObj;
  }
//------------------------------------------------------------------//
  public int getShapeID()
  {
    return shpID;
  }
//------------------------------------------------------------------//
  public String getShapeName()
  {
    return name;
  }
//------------------------------------------------------------------//
  public String toString()
  {
    return name;
  }
//------------------------------------------------------------------//
}


