package sfm.shp.network;

/**
 *  NodeFunction:  (1 / (1 + (Math.exp(-netInput))))  value-range:
 *   +36.7368006  == 1.0 output, 5 == 0.99
 *   0.0          == 0.5
 *   -36.7368006 == 1.1102229909599741E-16,  -5 = 0.0066
 */

import java.io.*;
import java.util.*;

import sfm.dataset.*;
import sfm.gui.GUI_SHAPES;
import sfm.gui.*;

public final class FNN implements Runnable, Serializable
{
  private static Vector<FNN> networkPool = new Vector<FNN>();
  private String netName;
  
  private Vector<Neuron[]> layerVector;
  
  private Neuron[] inputLayer, hiddenLayer, outputLayer;
  private Random rnd;
  
  private Vector<Instance> patterns_train, patterns_test, patterns_valset;
  private Dataset dataset;
  
  private double learningRate, momentum, maxError, noise;
  private int nInputNodes, nHiddenNodes, nOutputNodes, epochs;
  
  private ANNConfig annCfg;
  private GUI_SHELL gui;
  
  private double[] inputLayerNF, hiddenLayerNF;
  private int runCount, avgMSE_epochs, msePtr;
  private double cumMSE, minLR, maxLR_inc, inc;
  private double[] avgMSE;
//------------------------------------------------------------//
  /**
   * Constructor for a simple 3-layer feedforward 
   * neural network; input-, output- and 1 hidden layer.
   */
  public FNN(GUI_SHELL gooey, ANNConfig cfg, String name)
  {
    netName  = name;
    gui      = gooey;
    rnd      = new Random();
    maxError = cfg.getMargin();
    annCfg   = cfg; // initial default config
    
    patterns_train  = new Vector<Instance>();
    patterns_valset = new Vector<Instance>();
    patterns_test   = new Vector<Instance>();
    
    addNetwork();
  }
//-----------------------------------------------------------------//
  /**
   * Add new network to global pool, but *replace* if a network with 
   * the same name exists.
   */
  private void addNetwork()
  {
    for (int i=0; i < networkPool.size(); i++)
    {
      if (networkPool.elementAt(i).getNetworkName().equalsIgnoreCase(netName))
      {
        networkPool.setElementAt(this, i);
        return;
      }
    }
    networkPool.addElement(this);
  }
//-----------------------------------------------------------------//
  public static FNN getNetwork(String name)
  {
    for (int i=0; i < networkPool.size(); i++)
    {
      if (networkPool.elementAt(i).getNetworkName().equalsIgnoreCase(name))
      {
        return networkPool.elementAt(i);
      }
    }
    
    System.out.println("network " + name + " NOT FOUND!");
    return null;
  }
//-----------------------------------------------------------------//
  private String getNetworkName()
  {
    return netName;
  }
//-----------------------------------------------------------------//
  private void loadTrainingParameters()
  {
    epochs        = annCfg.getEpochs();
    learningRate  = annCfg.getLearningR();
    momentum      = annCfg.getMomentum();
    noise         = annCfg.getNoise();
    maxError      = annCfg.getMargin();
  }
//-----------------------------------------------------------------//
  private void loadConfig()
  {
    loadTrainingParameters();
    
    nInputNodes   = annCfg.getInputNodes();
    nHiddenNodes  = annCfg.getHiddenNodes();
    nOutputNodes  = annCfg.getOutputNodes();
    
    inputLayerNF   = new double[nInputNodes];
    hiddenLayerNF  = new double[nHiddenNodes];
    // realInputNodes = (nInputNodes - nHiddenNodes);
  }
//-----------------------------------------------------------------//
  public void setDataset(Dataset ds)
  {
    this.dataset = ds;
    
    dataset.splitDatasetAndStorePartitions(annCfg.getTrainingRatio(), 0.0);
    dataset.getPartitionedDatasets(patterns_train, patterns_valset, patterns_test);
  }
//-----------------------------------------------------------------//
  public void saveMe(String toFile) // Silje...
  {
    try
    {
      FileOutputStream fs    = new FileOutputStream(toFile + ".net");
      ObjectOutputStream out = new ObjectOutputStream(fs); 
      
      out.writeObject(annCfg);
      out.writeObject(inputLayer);
      out.writeObject(hiddenLayer);
      out.writeObject(outputLayer);
      
      out.flush();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
//-----------------------------------------------------------------//  
  public ANNConfig loadMe(String fromFile)
  {
    try
    {
      FileInputStream fs   = new FileInputStream(fromFile + ".net");
      ObjectInputStream in = new ObjectInputStream(fs); 
      
      annCfg      = (ANNConfig) in.readObject();
      inputLayer  = (Neuron[])  in.readObject();
      hiddenLayer = (Neuron[])  in.readObject();
      outputLayer = (Neuron[])  in.readObject();
      
      in.close();
      loadConfig(); // very crucial :)
      
      return annCfg;
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    catch (ClassNotFoundException e)
    {
      e.printStackTrace();
    }
    return null;
  }
//-----------------------------------------------------------------//
  public void createNetwork(ANNConfig cfg)
  {
    annCfg = cfg;
    loadConfig();
    
    inputLayer  = new Neuron[nInputNodes];
    hiddenLayer = new Neuron[nHiddenNodes];
    outputLayer = new Neuron[nOutputNodes];
    
    for (int i=0; i < nInputNodes;  i++) inputLayer[i]  = new Neuron(0, nHiddenNodes);
    for (int i=0; i < nHiddenNodes; i++) hiddenLayer[i] = new Neuron(nInputNodes, nOutputNodes);
    for (int i=0; i < nOutputNodes; i++) outputLayer[i] = new Neuron(nHiddenNodes, 0);
    
    layerVector = new Vector<Neuron[]>();
    layerVector.addElement(inputLayer);
    layerVector.addElement(hiddenLayer);
    layerVector.addElement(outputLayer);
    //---------------------------------------------------
    
    Link link;
    double weight;
    double flag;
    
    ///////////////////////////////////////////////
    /// FROM INPUT to HIDDEN LAYER    
    for (int i=0; i < inputLayer.length;  i++) 
    {
      for (int j=0; j < hiddenLayer.length; j++) 
      {
        flag = 1;
        if (Math.random() >= 0.5) flag = -1;        
        weight = rnd.nextDouble() * flag;      
        
        // weight = (rnd.nextDouble() * 2) - 1;
        link   = new Link(weight, false);
        
        link.setInNeuron(inputLayer[i]);
        link.setOutNeuron(hiddenLayer[j]);
        
        inputLayer[i].setForwardLink(j, link);        
        hiddenLayer[j].setBackwardLink(i, link);
      }
    }
    
    ///////////////////////////////////////////////
    /// FROM HIDDEN to OUTPUT LAYER
    rnd.setSeed(System.nanoTime());
    
    for (int i=0; i < hiddenLayer.length;  i++) 
    {
      for (int j=0; j < outputLayer.length; j++) 
      {
        flag = 1;
        if (Math.random() >= 0.5) flag = -1;        
        weight = rnd.nextDouble() * flag;
        
        // weight = (rnd.nextDouble() * 2) - 1;
        link   = new Link(weight, false);
        
        link.setInNeuron(hiddenLayer[i]);
        link.setOutNeuron(outputLayer[j]);
        
        hiddenLayer[i].setForwardLink(j, link);
        outputLayer[j].setBackwardLink(i, link);       
      }
    }
  }
//-----------------------------------------------------------------//
  /**
   * resets the input in the neurons to 0.0
   * Does not affect the INPUTLAYER, as it will implicitly be
   * reset when given a new input. 
   */
  private final void resetInput_HO()
  {
    for (int i=0; i < hiddenLayer.length; i++) hiddenLayer[i].setInput(0.0);
    for (int i=0; i < outputLayer.length; i++) outputLayer[i].setInput(0.0);
  }
//-----------------------------------------------------------------//
  private final void propagateInput()
  {
    for (int i=0; i < inputLayer.length; i++)  inputLayer[i].propagateForward();
    for (int i=0; i < hiddenLayer.length; i++) hiddenLayer[i].propagateForward();
  }
//-----------------------------------------------------------------//
  /**
   * for access by external callers
   */
  public final void runThisInputPattern(double[] pattern)
  {
    resetInput_HO();
    
    for (int i=0; i < pattern.length; i++)
    {
      inputLayer[i].setInput(pattern[i]);
    }
    propagateInput();
  }
//-----------------------------------------------------------------//
  private final void runThisInput(Instance inst)
  {
    double[] tmpPattern = inst.getPlainNeuralNetData();
    
    resetInput_HO();
    
    for (int i=0; i < tmpPattern.length; i++)
    {
      inputLayer[i].setInput(tmpPattern[i]);
    }
    propagateInput();
    
    // GUI_SHELL.printArray(inst.getClassID() + ": ", tmpPattern, 4);
    // GUI_SHELL.printArray("TT = ", inst.getTrainingTarget(), 8);
    // System.out.println("--------------------");
  }
//-----------------------------------------------------------------//
  /** 
   *  NB: Assuming FULL CONNECTIVITY, so that every node
   *  connects to ALL nodes in prev and/or next layer!
   */
  private final void computeInputAndHiddenLayerNFs()
  {
    for (int j=0; j < outputLayer[0].backwardConnections.length; j++)
    {
      hiddenLayerNF[j] = hiddenLayer[j].nodeFunction();
    }
    
    for (int j=0; j < hiddenLayer[0].backwardConnections.length; j++)
    {
      inputLayerNF[j] = inputLayer[j].nodeFunction();
    }
  }
//-----------------------------------------------------------------//
  
//-----------------------------------------------------------------//
  private final double backPropagate(Instance curPat)
  {
    double wChange, errorK = 0, errorJ = 0, outputK, outputJ, patternError = 0.0;
    double d, momentumAdd;
    int j, i;
    
    computeInputAndHiddenLayerNFs();
    
    double[] correctOutput = curPat.getTrainingTarget();
    
    // for each outputNode to hiddenNodes
    for (i=0; i < outputLayer.length; i++)
    {
      // if (i == correct) d = 1.0f; else d = 0.0; Not for this kindatraining...
      
      d = correctOutput[i];
      
      outputK = outputLayer[i].nodeFunction();
      errorK  =  outputK * (1.0f - outputK) * (d - outputK);
      
      // patternError += ((d - outputK) * (d - outputK));
      patternError += Math.abs(d - outputK);
      
      outputLayer[i].tmpError = errorK;
      
      // for each backwardlink in outputNode[i]... update the linkWeight back to hiddenLayer-nodes
      for (j=0; j < outputLayer[i].backwardConnections.length; j++)
      {
        // wChange = (learningRate * errorK * hiddenLayer[j].nodeFunction()); // the pure weight change
        wChange = (learningRate * errorK * hiddenLayerNF[j]); // the pure weight change
        
        momentumAdd = (momentum * outputLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff! 
        outputLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    
    // for each hiddenNode...to inputnodes
    for (i=0; i < hiddenLayer.length; i++)
    {
      outputJ = hiddenLayer[i].nodeFunction();
      errorJ  = (outputJ * (1 - outputJ)) * getSumOfWeightedErrors(hiddenLayer[i]);      
      
      for (j=0; j < hiddenLayer[i].backwardConnections.length; j++)
      {
        // wChange = (learningRate * (errorJ * inputLayer[j].nodeFunction())); // the pure weight change
        wChange = (learningRate * (errorJ * inputLayerNF[j])); // the pure weight change
        
        momentumAdd = (momentum * hiddenLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff!
        hiddenLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    return patternError;
  }
//-----------------------------------------------------------------//
  /**
   * implicit, usable only for the hiddenlayer - and only ONE hiddenlayer...
   * The previous layer must be the outputLayer 
   */
  private final double getSumOfWeightedErrors(Neuron n)
  {
    double error = 0.0f;
    
    for (int i=0; i < outputLayer.length; i++)
    {
      error += (n.forwardConnections[i].getWeight() * outputLayer[i].tmpError);
    }   
    return error;
  }
//--------------------------------------------------------------------//
  private final void adaptLearningRate(double epochError)
  {
    cumMSE += epochError;
    
    if (runCount++ == avgMSE_epochs)
    {
      runCount         = 0;
      avgMSE[msePtr++] = cumMSE; // avgMSE_epochs;
      cumMSE           = 0;
      
      if (msePtr == 2)
      {
        msePtr = 0;
        
        // INCREASE LR :: if epoch error is decreasing
        if (avgMSE[1] < avgMSE[0])
        {
          inc = learningRate * 0.1;
          momentum -= 0.005;
          
          if (inc > maxLR_inc)
          {
            inc = maxLR_inc;
          }
          
          if (momentum < 0.05) momentum = 0.05;
          
          learningRate += inc;
        }
        
        // INCREASE LR :: if epoch error is increasing
        else if (avgMSE[0] <= avgMSE[1])
        {
          learningRate -= (learningRate * 0.1);
          momentum += 0.01;
          
          // keep LR above a minimum, and also prevent a negative LR!
          if (learningRate < minLR)
          {
            learningRate = minLR;
          }
          
          // max momentum
          if (momentum > 4.0) momentum = 4.0;
        }
      }
    }
  }
//--------------------------------------------------------------------//
  private void startTrainingX()
  {
    gui.clearOutput();
    rnd.setSeed(System.currentTimeMillis());

    msePtr        = 0;
    runCount      = 0;    
    avgMSE_epochs = 3;
    
    cumMSE        = 0.0;
    minLR         = 0.00001;
    maxLR_inc     = 0.2;    
    avgMSE        = new double[2];
    
    double epochError = 0;
    int patternNo;
    Instance inst;
    
    Vector<Instance> tmpPatterns = new Vector<Instance>();
    // Vector<Double> stats         = new Vector<Double>();

    long startS = System.currentTimeMillis();
    ///////////////////////////////////////////////////////////
    for (int i=0; i < epochs; i++)
    {
      gui.showProgress(1);
      
      ////////////////////////////////////
      /// ADD random pattern-order! HERE:
      for (int j=0; j < patterns_train.size(); j++)
      {
        tmpPatterns.addElement(patterns_train.elementAt(j));
      }
      
      epochError = 0;
      
      while (!tmpPatterns.isEmpty())
      {
        patternNo = rnd.nextInt(tmpPatterns.size());
        inst = tmpPatterns.elementAt(patternNo);
        
        runThisInput(inst);        
        epochError += backPropagate(inst);
        
        tmpPatterns.remove(patternNo);
        
        if (annCfg.updateWeightsPerPattern()) updateWeights();
        // gui.showProgress(1);
      }
      
      if (annCfg.usingAdaptiveLearningRate()) adaptLearningRate(epochError);      
      if (annCfg.updateWeightsPerEpoch())     updateWeights();
      
      // stats.addElement(new Double(epochError / patterns_train.size()));       
      // if (isFinished()) break;
    }
    /////////////////////////////////////////////////////////     
    long endS = System.currentTimeMillis();
    
    printResultsX(((endS - startS) / 1000), patterns_train, epochError);
    gui.notify_finished();
    
    // System.out.println("final LR  : " + learningRate);
    // System.out.println("final MOM : " + momentum + "\n");    
  }
//--------------------------------------------------------------------//
  /**
   * Depending on the weight-updating scheme, updates the networks weights
   * either per pattern, or per epoch
   */
  private final void updateWeights()
  {
    for (int j=0; j < outputLayer.length; j++)  outputLayer[j].realizeBackwardLinks();
    for (int j=0; j < hiddenLayer.length; j++)  hiddenLayer[j].realizeBackwardLinks();
  }
//---------------------------------------------------------------------//
 /**
  * returns the total error; e.g. sum of error on all output units.
  * 
  * For a classification to be successful here, ALL output units
  * must have an error that is EQUAL-OR-LOWER than 'maxError'.
  */
  private boolean successful;
  
 private double testThisPattern(Instance p)
 {
   runThisInput(p);
      
   double totalError = 0.0, value, diff;
   double[] correctValues = p.getTrainingTarget(); 
   
   successful = true; // assume success...
   
   for (int i=0; i < outputLayer.length; i++)
   {
     value = outputLayer[i].nodeFunction();

     diff = Math.abs((value - correctValues[i]));
     totalError += diff;
     
     if (diff > maxError) successful = false;
   }
   
   return totalError;
 }
//---------------------------------------------------------------------//
 private void printResultsX(long secs, Vector<Instance> patterns, double epochError)
 {
   ///////////////////////////////////////////////////////////
   ///// run through all patterns again,  with no training
   int successes = 0;
   double patternError;
   Instance instance;

   gui.clearOutput();
   gui.print("LIST OF FAILED PATTERN CLASSIFICATIONS:");
   
   for (int i=0; i < patterns.size(); i++)
   {
     instance     = patterns.elementAt(i);      
     patternError = testThisPattern(instance);
     
     if (successful) // ALL outputs were within error threshold
     {
       successes++;
     }
     else
     {
       gui.print("\n" + instance.getLabel() + " failed.    Error = " + patternError);
     }
   }
   
   gui.print("\n------------------------------------");
   
   gui.print("\n\nMSE for last epoch : " + (epochError / patterns.size()));
   gui.print("\nEpochs run         : " + gui.getEpochsRunInCurrentSession());
   gui.print("\nMax error allowed  : " + maxError);
   gui.print("\nTime elapsed       : " + secs + " secs");
   
   double pct = ((double)successes / (double)patterns.size()) * 100.0;
   gui.print("\nACCURACY : " + pct + "%   [ " + successes + " of " + patterns.size() + " ]");
 }
//--------------------------------------------------------------------//
 public Vector<Instance> getTrainingData()
 {
   return this.patterns_train;
 }
//--------------------------------------------------------------------//
 public Vector<Instance> getTestData()
 {
   return this.patterns_test;
 }
//--------------------------------------------------------------------//
 public int getNumberOfTrainingSamples()
 {
   return patterns_train.size();
 }
//-------------------------------------------------------------------//
 public int getNumberOfTestingSamples()
 {
   return patterns_test.size();
 }
//---------------------------------------------------------------------//
 /**
  * In case of multiple hiddenlayers, the last one is returned!
  */
 public Neuron[] getHiddenLayer()
 { 
   return layerVector.elementAt(layerVector.size()-2);
 }
//---------------------------------------------------------------------//
 public Neuron[] getOutputLayer()
 { 
   return outputLayer;
 }
//---------------------------------------------------------------------//
 public double[] getLayerActivation(Neuron[] layer)
 {
   double[] actualOutput = new double[layer.length];
   
   for (int i=0; i < actualOutput.length; i++)
   {
     actualOutput[i] = layer[i].nodeFunction();
   }
   return actualOutput;   
 }
//---------------------------------------------------------------------//
 public void startTraining(ANNConfig cfg) 
 {
   annCfg = cfg;
   loadTrainingParameters();
   
   Thread t = new Thread(this);
   t.setPriority(Thread.NORM_PRIORITY);
   t.start();
 }
//--------------------------------------------------------------------//   
 public void startTesting()
 {
   // yo !?
   
   printResultsX(0, patterns_test, 0.0);
 }   
//---------------------------------------------------------------------//   
 public void run()
 {
   startTrainingX();   
 }
//---------------------------------------------------------------------//  
}